import logging

from lettrade.account import LetAccountInsufficientException
from lettrade.commander import Commander
from lettrade.data import DataFeed, DataFeeder, LetNoMoreDataFeedException
from lettrade.exchange import (
    Exchange,
    Execution,
    LetOrderValidateException,
    Order,
    Position,
)
from lettrade.strategy import Strategy

logger = logging.getLogger(__name__)


class Brain:
    """Brain of bot"""

    strategy: Strategy
    exchange: Exchange
    feeder: DataFeeder
    commander: Commander

    # datas: list[DataFeed]
    data: DataFeed

    def __init__(
        self,
        strategy: Strategy,
        exchange: Exchange,
        feeder: DataFeeder,
        commander: Commander,
        **kwargs,
    ) -> None:
        """_summary_

        Args:
            strategy (Strategy): _description_
            exchange (Exchange): _description_
            feeder (DataFeeder): _description_
            commander (Commander): _description_
        """
        self.strategy = strategy
        self.exchange = exchange
        self.feeder = feeder
        self.commander = commander

        # self.datas = self.feeder.datas
        self.data = self.feeder.data

    def start(
        self,
        feed_size: int = 0,
        **kwargs,
    ):
        """_summary_

        Args:
            feed_size (int, optional): Init at main datafeeds size. Defaults to 1.
        """
        self.data._set_main()
        self.strategy.init()

        # Load init data
        self.feeder.start(size=feed_size)

        # Start exchange
        self.exchange.start()

        # Start strategy
        self.strategy._start()

    def run(self):
        """Run the trading bot"""

        while self.feeder.alive():
            # Load feeder next data
            try:
                self.feeder.next()
                self.exchange.next()
                self.strategy._next()
                self.exchange.next_next()
            except LetOrderValidateException as e:
                logger.error(
                    "[%s] Order validates exception",
                    self.data.now,
                    exc_info=e,
                )
                continue
            except LetAccountInsufficientException as e:
                logger.error("Account equity is insufficient", exc_info=e)
                break
            except LetNoMoreDataFeedException:
                break
            except Exception as e:
                logger.exception("Bot running error", exc_info=e)
                break

        self.strategy._stop()

    def stop(self):
        """Stop the trading bot"""
        self.feeder.stop()
        self.exchange.stop()

    # Events
    def on_executions(self, executions: list[Execution]):
        """Receive new `Execution` event and send to `Strategy`"""
        self.on_transactions(executions)
        self.strategy.on_executions(executions)

    def on_orders(self, orders: list[Order]):
        """Receive new `Order` event and send to `Strategy`"""
        self.on_transactions(orders)
        self.strategy.on_orders(orders)

    def on_positions(self, positions: list[Position]):
        """Receive new `Position` event and send to `Strategy`"""
        self.on_transactions(positions)
        self.strategy.on_positions(positions)

    def on_notify(self, *args, **kwargs):
        """Receive new notify and send to Strategy"""
        self.strategy.on_notify(*args, **kwargs)

    def on_transactions(self, transactions):
        """Receive new transaction event and send to `Strategy`"""
        if self.commander is not None:
            # TODO: send message to commander when new transaction
            self.commander.send_message(f"New transactions: {str(transactions)}")

        self.strategy.on_transactions(transactions)



from typing import Any, Literal

import numpy as np
import pandas as pd


def pandas_inject(obj: object | None = None):
    if obj is None:
        from pandas.core.base import PandasObject

        obj = PandasObject

    obj.signal_direction = signal_direction
    obj.signal_condiction = signal_condiction


def signal_direction(
    dataframe: pd.DataFrame,
    up: pd.Series,
    down: pd.Series,
    name: str,
    value: int = 0,
    value_up: int = 100,
    value_down: int = -100,
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_type: Literal["line", "mark"] = "line",
    plot_kwargs: dict | None = None,
    **kwargs,
) -> pd.Series | pd.DataFrame:
    """Define a signal with 2 direction Up and Down with fixed value

    Args:
        dataframe (pd.DataFrame): _description_
        up (pd.Series): _description_
        down (pd.Series): _description_
        name (str): Name of signal, column name when add to DataFrame with inplace=True.
        value (int, optional): Default value when condiction is not matched. Defaults to 0.
        value_up (int, optional): _description_. Defaults to 100.
        value_down (int, optional): _description_. Defaults to -100.
        inplace (bool, optional): Whether to add to the DataFrame and return DataFrame rather than return result. Defaults to False.
        plot (bool | list, optional): _description_. Defaults to False.
        plot_type (Literal[&quot;line&quot;, &quot;mark&quot;], optional): _description_. Defaults to "line".
        plot_kwargs (dict | None, optional): _description_. Defaults to None.

    Returns:
        pd.Series | pd.DataFrame: _description_
    """
    return signal_condiction(
        dataframe,
        [up, value_up],
        [down, value_down],
        name=name,
        value=value,
        inplace=inplace,
        plot=plot,
        plot_type=plot_type,
        plot_kwargs=plot_kwargs,
        **kwargs,
    )


def signal_condiction(
    dataframe: pd.DataFrame,
    *condictions: list[list[pd.Series | Any]],
    name: str,
    value: int | float = np.nan,
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_type: Literal["line", "mark"] = "line",
    plot_kwargs: dict | None = None,
    **kwargs,
) -> pd.Series | pd.DataFrame:
    """Define a signal with multiple condiction

    Args:
        dataframe (pd.DataFrame): _description_
        *condictions (list[list[pd.Series | Any]]): Pairs of condiction [`<pandas.Series condiction>`, `<value>`]
        name (str): Name of signal, column name when add to DataFrame with inplace=True.
        value (int, optional): Default value when condiction is not matched. Defaults to 0.
        inplace (bool, optional): _description_. Defaults to False.
        plot (bool | list, optional): _description_. Defaults to False.
        plot_type (Literal[&quot;line&quot;, &quot;mark&quot;], optional): _description_. Defaults to "line".
        plot_kwargs (dict | None, optional): _description_. Defaults to None.

    Usage:
        ```python
        df.i.signal_condiction(
            [df["close"] > df["open"], 100],
            [df["close"] < df["open"], -100],
            name="cdl_direction",
            inplace=True,
            plot=True,
            plot_kwargs=dict(color="green", width=5),
        )
        ```

    Raises:
        RuntimeError: _description_

    Returns:
        pd.Series | pd.DataFrame: _description_
    """
    if __debug__:
        if not isinstance(dataframe, pd.DataFrame):
            raise RuntimeError(
                f"dataframe type '{type(dataframe)}' "
                "is not instance of pandas.DataFrame"
            )
        if plot and not inplace:
            raise RuntimeError("Cannot plot when inplace=False")

    s = pd.Series(value, index=dataframe.index, name=name, **kwargs)
    for condiction in condictions:
        s.loc[condiction[0]] = pd.to_numeric(condiction[1], errors='coerce')

    if inplace:
        dataframe[name] = s

        # Plot
        if plot:
            if plot_kwargs is None:
                plot_kwargs = dict()

            plot_kwargs.update(series=name, name=name)

            from lettrade.indicator.plot import IndicatorPlotter
            from lettrade.plot.plotly import plot_line, plot_mark

            plotter = plot_mark if plot_type == "mark" else plot_line
            IndicatorPlotter(dataframe=dataframe, plotter=plotter, **plot_kwargs)

        return dataframe

    return s



def plot_merge(
    source: dict,
    *updates: list[dict],
    recursive: int = 0,
    recursive_max: int = 1,
) -> dict:
    """Merge multiple update plot config to source config

    Args:
        source (dict): _description_
        *updates (list[dict]): _description_

    Returns:
        dict: Merged config
    """
    for update in updates:
        for key, value in update.items():
            if key not in source:
                source[key] = value
                continue

            if isinstance(source[key], list):
                source[key].extend(value)
                continue
            if isinstance(source[key], dict):
                if recursive >= recursive_max:
                    source[key] = value
                    continue

                sub = source[key]
                plot_merge(
                    sub,
                    value,
                    recursive=recursive + 1,
                    recursive_max=recursive_max,
                )
                continue
    return source



import logging
from abc import ABC, abstractmethod
from datetime import datetime
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .exchange import LiveExchange
    from .trade import LiveOrder, LivePosition

logger = logging.getLogger(__name__)


class LiveAPI(ABC):
    @classmethod
    def multiprocess(cls, kwargs: dict, **other_kwargs):
        """"""

    def __init__(self, **kwargs):
        """"""

    def init(self, **kwargs):
        """"""

    def start(self, exchange: "LiveExchange"):
        """"""

    def next(self):
        """"""

    def stop(self):
        """"""

    # Public
    def heartbeat(self) -> bool:
        """"""
        return True

    # Market
    @abstractmethod
    def market(self, symbol: str) -> dict:
        """"""

    @abstractmethod
    def markets(self, symbols: list[str]) -> dict:
        """"""

    # Bars
    @abstractmethod
    def bars(
        self,
        symbol,
        timeframe,
        since: int | datetime | None = 0,
        to: int | datetime | None = 1_000,
    ) -> list[list]:
        """"""

    # Tick
    @abstractmethod
    def tick_get(self, symbol: str) -> dict:
        """"""

    ### Private
    # Account
    @abstractmethod
    def account(self) -> dict:
        """"""

    #  Order
    @abstractmethod
    def orders_total(
        self,
        since: datetime | None = None,
        to: datetime | None = None,
        **kwargs,
    ) -> int:
        """"""

    @abstractmethod
    def orders_get(
        self,
        id: str | None = None,
        symbol: str | None = None,
        **kwargs,
    ):
        """"""

    @abstractmethod
    def order_open(self, **kwargs) -> dict:
        """"""

    @abstractmethod
    def order_update(self, order: "LiveOrder", sl=None, tp=None, **kwargs) -> dict:
        """"""

    @abstractmethod
    def order_close(self, order: "LiveOrder", **kwargs) -> dict:
        """"""

    def orders_history_get(
        self,
        id: str | None = None,
        since: datetime | None = None,
        to: datetime | None = None,
        **kwargs,
    ) -> list[dict]:
        """"""

    # Execution
    @abstractmethod
    def executions_total(
        self,
        since: datetime | None = None,
        to: datetime | None = None,
        **kwargs,
    ) -> int:
        """"""

    @abstractmethod
    def executions_get(
        self,
        position_id: str | None = None,
        search: str | None = None,
        **kwargs,
    ) -> list[dict]:
        """"""

    @abstractmethod
    def execution_get(self, id: str, **kwargs) -> dict:
        """"""

    # Position
    @abstractmethod
    def positions_total(
        self,
        since: datetime | None = None,
        to: datetime | None = None,
        **kwargs,
    ) -> int:
        """"""

    @abstractmethod
    def positions_get(self, id: str = None, symbol: str = None, **kwargs) -> list[dict]:
        """"""

    @abstractmethod
    def position_update(
        self,
        position: "LivePosition",
        sl: float | None = None,
        tp: float | None = None,
        **kwargs,
    ) -> dict:
        """"""

    @abstractmethod
    def position_close(self, position: "LivePosition", **kwargs) -> dict:
        """"""



import logging
from datetime import datetime

import pandas as pd

from lettrade.data import DataFeed

from .api import LiveAPI

logger = logging.getLogger(__name__)


class LiveDataFeed(DataFeed):
    """Live trading DataFeed"""

    _api_cls: type[LiveAPI] = LiveAPI
    _bar_datetime_unit: str = "ms"

    def __init__(
        self,
        symbol: str,
        timeframe: str | int | pd.Timedelta,
        name: str | None = None,
        columns: list[str] | None = None,
        api: LiveAPI | None = None,
        api_kwargs: dict | None = None,
        **kwargs,
    ) -> None:
        """_summary_

        Args:
            symbol (str): Symbol of DataFeed
            timeframe (str | int | pd.Timedelta): TimeFrame of DataFeed
            name (str | None, optional): Name of DataFeed, auto generate `{symbol}_{timeframe}` if none. Defaults to None.
            columns (list[str] | None, optional): List of DataFeed columns. Defaults to None.
            api (LiveAPI | None, optional): Live trading API. Defaults to None.
        """
        super().__init__(
            name=name or f"{symbol}_{timeframe}",
            timeframe=timeframe,
            columns=columns or ["open", "high", "low", "close", "volume"],
            **kwargs,
        )

        self.meta.update(symbol=symbol, base_columns=self.columns.copy())

        if api is not None:
            self._api = api
        elif api_kwargs is not None:
            self._api = self._api_cls(**api_kwargs)
        else:
            raise RuntimeError("Parameter: api or api_kwargs cannot missing")

    # Properties
    @property
    def symbol(self) -> str:
        """Property to get symbol of DataFeed"""
        return self.meta["symbol"]

    @property
    def _base_columns(self) -> LiveAPI:
        return self.meta["base_columns"]

    @property
    def _api(self) -> LiveAPI:
        return getattr(self, "__api")

    @_api.setter
    def _api(self, value) -> LiveAPI:
        object.__setattr__(self, "__api", value)

    # Functions
    def symbol_info(self):
        """Get symbol information from API"""
        return self._api.market(symbol=self.symbol)

    def copy(self, deep: bool = False, **kwargs) -> DataFeed:
        return super().copy(deep, symbol=self.symbol, **kwargs)

    def next(self, size=1, tick=0) -> bool:
        """Drop extra columns and load next DataFeed

        Args:
            size (int, optional): _description_. Defaults to 1.
            tick (int, optional): _description_. Defaults to 0.

        Returns:
            bool: _description_
        """
        # Drop existed extra columns to skip reusing calculated data
        self.drop(columns=self.columns.difference(self._base_columns), inplace=True)

        self.bars_load(since=0, to=size + 1)
        self.l.go_stop()
        return True

    def bars_load(
        self,
        since: int | str | pd.Timestamp,
        to: int | str | pd.Timestamp,
    ) -> bool:
        """Get bar from API and push to DataFeed

        Args:
            since (int | str | pd.Timestamp): _description_
            to (int | str | pd.Timestamp): _description_

        Returns:
            bool: True if has data, False if no data
        """
        bars = self.bars(since=since, to=to)

        if bars is None or len(bars) == 0:
            logger.warning("No bars data for %s", self.name)
            return False

        self.push(bars, unit=self._bar_datetime_unit)
        return True

    def bars(
        self,
        since: int | str | pd.Timestamp,
        to: int | str | pd.Timestamp,
    ) -> list:
        """Get bars from LiveAPI

        Args:
            since (int | str | pd.Timestamp): _description_
            to (int | str | pd.Timestamp): _description_

        Returns:
            list: list of bar
        """
        return self._api.bars(
            symbol=self.symbol,
            timeframe=self.timeframe.string,
            since=since,
            to=to,
        )

    ### Extend
    def dump_csv(
        self,
        path: str | None = None,
        since: int | str | datetime | None = 0,
        to: int | str | datetime | None = 1_000,
        **kwargs,
    ):
        """_summary_

        Args:
            path (str | None, optional): _description_. Defaults to None.
            since (int  |  str  |  datetime | None, optional): _description_. Defaults to 0.
            to (int  |  str  |  datetime | None, optional): _description_. Defaults to 1_000.
        """
        if self.empty:
            if isinstance(since, str):
                since = pd.to_datetime(since).to_pydatetime()
            if isinstance(to, str):
                to = pd.to_datetime(to).to_pydatetime()

            self.bars_load(since=since, to=to)

        if path is None:
            path = f"data/{self.name}-{since}_{to}.csv"

        from lettrade.data.extra.csv import csv_export

        csv_export(dataframe=self, path=path, **kwargs)

    # @classmethod
    # def instance(
    #     cls,
    #     api: LiveAPI | None = None,
    #     api_kwargs: dict | None = None,
    #     **kwargs,
    # ) -> "LiveDataFeed":
    #     """_summary_

    #     Args:
    #         api (LiveAPI, optional): _description_. Defaults to None.
    #         api_kwargs (dict, optional): _description_. Defaults to None.

    #     Raises:
    #         RuntimeError: Missing api requirement

    #     Returns:
    #         LiveDataFeed: DataFeed object
    #     """
    #     if api is None:
    #         if api_kwargs is None:
    #             raise RuntimeError("api or api_kwargs cannot missing")
    #         api = cls._api_cls(**api_kwargs)
    #     obj = cls(api=api, **kwargs)
    #     return obj



import logging

import numpy as np
import pandas as pd

from lettrade.account import Account
from lettrade.data import DataFeeder, TimeFrame
from lettrade.exchange import Exchange
from lettrade.strategy import Strategy

logger = logging.getLogger(__name__)


class BotStatistic:
    """
    Compute strategy result
    """

    result: pd.Series

    def __init__(
        self,
        feeder: DataFeeder,
        exchange: Exchange,
        strategy: Strategy,
    ) -> None:
        self.feeder: DataFeeder = feeder
        self.exchange: Exchange = exchange
        self.strategy: Strategy = strategy
        self.account: Account = strategy.account

    def stop(self):
        pass

    def compute(self):
        """Calculate strategy report"""
        data = self.feeder.data

        ### Stats
        self.result = result = pd.Series(dtype=object)

        result.loc["strategy"] = str(self.strategy.__class__)
        result.loc["start"] = data.index[0]
        result.loc["end"] = data.index[-1]
        result.loc["duration"] = result.end - result.start

        ### Equity
        equities_index = pd.DatetimeIndex(self.account._equities.keys())
        equities = pd.Series(self.account._equities.values(), index=equities_index)
        dd = 1 - equities / np.maximum.accumulate(equities)
        dd_dur, dd_peaks = _compute_drawdown_duration_peaks(
            pd.Series(dd, index=equities_index)
        )

        result.loc["start_balance"] = round(equities.iloc[0], 2)
        result.loc["equity"] = round(equities.iloc[-1], 2)
        result.loc["equity_peak"] = round(equities.max(), 2)

        pl = equities.iloc[-1] - equities.iloc[0]
        result.loc["pl"] = round(pl, 2)
        result.loc["pl_percent"] = round(pl / equities.iloc[0] * 100, 2)

        c = data.close.values
        result.loc["buy_hold_pl_percent"] = round(
            (c[-1] - c[0]) / c[0] * 100, 2
        )  # long-only return

        max_dd = -np.nan_to_num(dd.max())
        result.loc["max_drawdown_percent"] = round(max_dd * 100, 2)
        result.loc["avg_drawdown_percent"] = round(-dd_peaks.mean() * 100, 2)
        result.loc["max_drawdown_duration"] = _round_timedelta(
            dd_dur.max(), data.timeframe
        )
        result.loc["avg_drawdown_duration"] = _round_timedelta(
            dd_dur.mean(), data.timeframe
        )

        # Separator
        result.loc[""] = ""

        ### Positions
        positions = list(self.exchange.history_positions.values()) + list(
            self.exchange.positions.values()
        )
        positions_columns = (
            "size",
            "entry_at",
            "exit_at",
            "entry_price",
            "exit_price",
            "pl",
            "fee",
        )
        positions_df = pd.DataFrame(columns=positions_columns)
        for position in positions:
            positions_df.at[position.id, positions_columns] = (
                position.size,
                position.entry_at,
                position.exit_at,
                position.entry_price,
                position.exit_price,
                position.pl,
                position.fee,
            )
        positions_df["duration"] = positions_df["entry_at"] - positions_df["exit_at"]
        positions_total = len(positions)
        pl = positions_df["pl"]

        result.loc["positions"] = positions_total

        win_rate = np.nan if not positions_total else (pl > 0).mean()
        result.loc["win_rate"] = round(win_rate, 2)
        result.loc["fee"] = round(positions_df.fee.sum(), 2)
        result.loc["best_trade_percent"] = round(pl.max(), 2)
        result.loc["worst_trade_percent"] = round(pl.min(), 2)
        result.loc["sqn"] = round(
            np.sqrt(positions_total) * pl.mean() / (pl.std() or np.nan),
            2,
        )
        result.loc["kelly_criterion"] = win_rate - (1 - win_rate) / (
            pl[pl > 0].mean() / -pl[pl < 0].mean()
        )
        result.loc["profit_factor"] = pl[pl > 0].sum() / (
            abs(pl[pl < 0].sum()) or np.nan
        )

        return self.result

    def __repr__(self) -> str:
        result = self.result.rename(
            {
                "strategy": "# Strategy",
                "start": "Start",
                "end": "End",
                "duration": "Duration",
                "start_balance": "Start Balance",
                "equity": "Equity [$]",
                "equity_peak": "Equity Peak [$]",
                "pl": "PL [$]",
                "pl_percent": "PL [%]",
                "buy_hold_pl_percent": "Buy & Hold PL [%]",
                "max_drawdown_percent": "Max. Drawdown [%]",
                "avg_drawdown_percent": "Avg. Drawdown [%]",
                "max_drawdown_duration": "Max. Drawdown Duration",
                "avg_drawdown_duration": "Avg. Drawdown Duration",
                "positions": "# Positions",
                "win_rate": "Win Rate [%]",
                "fee": "Fee [$]",
                "best_trade_percent": "Best Trade [%]",
                "worst_trade_percent": "Worst Trade [%]",
                "profit_factor": "Profit Factor",
                "kelly_criterion": "Kelly Criterion",
                "sqn": "SQN",
            }
        )
        return str(result.to_string())

    def show(self):
        """
        Show statistic report
        """
        if "Start" not in self.result:
            logger.warning("call compute() before show()")
            self.compute()

        # Show result inside docs session
        if __debug__:
            from lettrade.utils.docs import is_docs_session

            if is_docs_session():
                print(str(self))
                return

        logger.info(
            "\n============= Statistic result =============\n%s\n",
            str(self),
        )


def _round_timedelta(value, timeframe: TimeFrame):
    if not isinstance(value, pd.Timedelta):
        return value
    return timeframe.ceil(value)


def _compute_drawdown_duration_peaks(dd: pd.Series):
    # TODO: clean
    iloc = np.unique(np.r_[(dd == 0).values.nonzero()[0], len(dd) - 1])
    iloc = pd.Series(iloc, index=dd.index[iloc])
    df = iloc.to_frame("iloc").assign(prev=iloc.shift())
    df = df[df["iloc"] > df["prev"] + 1].astype(int)

    # If no drawdown since no trade, avoid below for pandas sake and return nan series
    if not len(df):
        return (dd.replace(0, np.nan),) * 2

    df["duration"] = df["iloc"].map(dd.index.__getitem__) - df["prev"].map(
        dd.index.__getitem__
    )
    df["peak_dd"] = df.apply(
        lambda row: dd.iloc[row["prev"] : row["iloc"] + 1].max(), axis=1
    )
    df = df.reindex(dd.index)
    return df["duration"], df["peak_dd"]



from .account import *
from .brain import *
from .commander import *
from .data import *
from .exchange import *
from .lettrade import LetTrade, LetTradeBot
from .plot import *
from .stats import *
from .strategy import *



import logging
import re
from typing import TYPE_CHECKING, Callable

import pandas as pd

from .timeframe import TimeFrame
from .wrapper import LetDataFeedWrapper

if TYPE_CHECKING:
    from lettrade import indicator

logger = logging.getLogger(__name__)

_data_name_pattern = re.compile(r"^[\w\_]+$")


class DataFeed(pd.DataFrame):
    """Data for Strategy. A implement of pandas.DataFrame"""

    l: LetDataFeedWrapper
    """LetTrade DataFeed wrapper using to manage index pointer of DataFeed"""

    def __init__(
        self,
        *args,
        name: str,
        timeframe: TimeFrame,
        meta: dict | None = None,
        **kwargs,
    ) -> None:
        """_summary_

        Args:
            name (str): _description_
            timeframe (TimeFrame): _description_
            meta (dict | None, optional): _description_. Defaults to None.

        Raises:
            RuntimeError: _description_
        """
        # Validate
        if not _data_name_pattern.match(name):
            raise RuntimeError(
                f"Bot name {name} is not valid format {_data_name_pattern}"
            )

        # Init
        super().__init__(*args, **kwargs)
        self._init_index()

        # Metadata
        if not meta:
            meta = dict()
        meta["name"] = name
        meta["timeframe"] = TimeFrame(timeframe)
        self.attrs = {"lt_meta": meta}

        # LetWrapper
        object.__setattr__(self, "l", LetDataFeedWrapper(self))

    def __setstate__(self, data):
        super().__setstate__(data)
        if not hasattr(self, "l"):
            object.__setattr__(self, "l", LetDataFeedWrapper(self))

    # Internal
    def _init_index(self):
        if not isinstance(self.index, pd.DatetimeIndex):
            if not self.empty:
                raise RuntimeError("Index is not pandas.DatetimeIndex format")
            self.index = self.index.astype("datetime64[ns, UTC]")

        self.index.rename("datetime", inplace=True)

    def _set_main(self):
        """Set this dataframe is main datafeed"""
        self.meta["is_main"] = True

    # External
    def copy(self, deep: bool = False, **kwargs) -> "DataFeed":
        """_summary_

        Args:
            deep (bool, optional): _description_. Defaults to False.

        Returns:
            DataFeed: _description_
        """
        df = super().copy(deep=deep)
        df = self.__class__(
            data=df,
            name=self.name,
            timeframe=self.timeframe,
            meta=self.meta.copy(),
            **kwargs,
        )
        return df

    def next(self, size=1):
        """Load next data

        Args:
            size (int, optional): _description_. Defaults to 1.
        """
        self.l.next(size)

    def bar(self, i: int = 0) -> pd.Timestamp:
        """Get current pd.Timestamp value of DataFeed

        Args:
            i (int, optional): Index. Defaults to 0.

        Returns:
            pd.Timestamp: _description_
        """
        return self.l.index[i]

    def push(
        self,
        rows: list[list[int | float]],
        unit: str | None = None,
        utc: bool = True,
        **kwargs,
    ):
        """Push new rows to DataFeed

        Args:
            rows (list[list[int | float]]): list of rows `[["timestamp", "open price", "high price"...]]`
            unit (str | None, optional): pandas.Timestamp parsing unit. Defaults to None.
            utc (bool, optional): _description_. Defaults to True.
        """
        for row in rows:
            dt = pd.to_datetime(row[0], unit=unit, utc=utc, **kwargs)
            self.at[
                dt,
                (
                    "open",
                    "high",
                    "low",
                    "close",
                    "volume",
                ),
            ] = (
                row[1],  # open
                row[2],  # high
                row[3],  # low
                row[4],  # close
                row[5],  # volume
            )

        if __debug__:
            logger.debug("[%s] Update bar: \n%s", self.name, self.tail(len(rows)))

    def drop(
        self,
        *args,
        since: int | str | pd.Timestamp | None = None,
        to: int | str | pd.Timestamp | None = None,
        **kwargs,
    ) -> None:
        """_summary_

        Args:
            since (int | str | pd.Timestamp | None, optional): _description_. Defaults to None.
            to (int | str | pd.Timestamp | None, optional): _description_. Defaults to None.

        Raises:
            RuntimeError: _description_
        """
        if since is None and to is None:
            super().drop(*args, **kwargs)
            return

        condiction = None

        # Since
        if since is not None:
            if isinstance(since, int):
                loc = self.l.index[since]
            elif isinstance(since, str):
                loc = pd.to_datetime(since, utc=True)
            elif isinstance(since, pd.Timestamp):
                loc = since
            else:
                raise RuntimeError(f"DataFeed.drop since {since} is invalid")
            condiction = self.index < loc

        # To
        if to is not None:
            if isinstance(to, int):
                loc = self.l.index[to]
            elif isinstance(to, str):
                loc = pd.to_datetime(to, utc=True)
            elif isinstance(to, pd.Timestamp):
                loc = to
            else:
                raise RuntimeError(f"DataFeed.drop to {to} is invalid")

            if condiction is None:
                condiction = self.index > loc
            else:
                condiction = condiction | (self.index > loc)

        index = self[condiction].index
        super().drop(index=index, inplace=True)
        self.l.reset()

        if __debug__:
            logger.debug("BackTestDataFeed %s dropped %s rows", self.name, len(index))

    @property
    def now(self) -> pd.Timestamp:
        """Property to get current index value of DataFeed"""
        return self.l.index[0]

    @property
    def meta(self) -> dict:
        """Property to get metadata of DataFeed"""
        return self.attrs["lt_meta"]

    @property
    def name(self) -> str:
        """Property to get name of DataFeed"""
        return self.meta["name"]

    @property
    def timeframe(self) -> TimeFrame:
        """Property to get timeframe of DataFeed"""
        return self.meta["timeframe"]

    @property
    def is_main(self) -> bool:
        """Property to check DataFeed is main DataFeed or not"""
        return self.meta.get("is_main", False)

    @property
    def i(self) -> "indicator":
        """Alias to `lettrade.indicator` and using in DataFeed by call: `DataFeed.i.indicator_name()`"""
        if not hasattr(self, "_lt_indicators_injector"):
            object.__setattr__(
                self,
                "_lt_indicators_injector",
                _LetIndicatorsInjector(self),
            )
        return self._lt_indicators_injector


class _LetIndicatorsInjector:
    """Class help to inject indicator into pandas.DataFrame with prefix `_lt_i_`.
    Avoid conflic between pandas.DataFrame column name and indicator function name
    """

    def __init__(self, df: DataFeed) -> None:
        self.__dict__["df"] = df

        from lettrade import indicator

        indicator.indicators_inject_pandas(self)

    def __setattr__(self, name: str, value: Callable) -> None:
        from pandas.core.base import PandasObject

        i_name = f"_lt_i_{name}"
        if hasattr(PandasObject, i_name):
            raise RuntimeError(f"Indicator {i_name} existed")

        setattr(PandasObject, i_name, value)

    def __getattr__(self, name: str) -> Callable:
        return getattr(self.__dict__["df"], f"_lt_i_{name}")


if __debug__:
    # __debug__ code willbe remove when run in python production flag `python -O` or `python -OO`
    from lettrade.base.flag import validate_data_getitem_pointer

    if validate_data_getitem_pointer:
        # Check missing data get item by pointer:
        # Wrong: data.index[<pointer>] -> Right: data.l.index[<pointer>]
        # Wrong: data.open[<pointer>] -> Right: data.l.open[<pointer>]

        import logging

        logger = logging.getLogger(__name__)

        def inject_data_validator(cls):
            # __class__ = cls

            def data_getitem(self, value):
                if isinstance(value, int):
                    logger.warning(
                        "[%s] Get data by pointer %d many wrong data. "
                        "using <data>.l[<pointer>] to get data at pointer",
                        self.__class__,
                        value,
                    )
                return self.__lt__getitem__(value)

            cls.__lt__getitem__ = cls.__getitem__
            cls.__getitem__ = data_getitem

        inject_data_validator(DataFeed)
        inject_data_validator(pd.Series)
        inject_data_validator(pd.DatetimeIndex)



import pandas as pd
import talib.abstract as ta


def atr(
    dataframe: pd.DataFrame,
    window: int = 20,
    name: str | None = None,
    prefix: str = "",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_kwargs: dict | None = None,
) -> dict[str, pd.Series] | pd.DataFrame:
    """_summary_

    Args:
        dataframe (pd.DataFrame): _description_
        window (int, optional): _description_. Defaults to 20.
        name (str | None, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "".
        inplace (bool, optional): _description_. Defaults to False.
        plot (bool | list, optional): _description_. Defaults to False.
        plot_kwargs (dict | None, optional): _description_. Defaults to None.

    Raises:
        RuntimeError: _description_

    Returns:
        dict[str, pd.Series] | pd.DataFrame: {atr}
    """
    if __debug__:
        if not isinstance(dataframe, pd.DataFrame):
            raise RuntimeError(
                f"dataframe type '{type(dataframe)}' "
                "is not instance of pandas.DataFrame"
            )
        if plot and not inplace:
            raise RuntimeError("Cannot plot when inplace=False")

    i = ta.ATR(dataframe, timeperiod=window)

    if inplace:
        name = name or f"{prefix}atr"
        dataframe[name] = i

        # Plot
        if plot:
            if plot_kwargs is None:
                plot_kwargs = dict()

            plot_kwargs.update(series=name, name=name)
            plot_kwargs.setdefault("row", 2)
            plot_kwargs.setdefault("row_height", 0.5)

            from lettrade.indicator.plot import IndicatorPlotter
            from lettrade.plot.plotly import plot_line

            IndicatorPlotter(dataframe=dataframe, plotter=plot_line, **plot_kwargs)

        return dataframe

    return i



"""Module help LetTrade send notify and receive command from Telegram Bot

# Example

Example:
    ```python
    --8<-- "example/commander/telegram_mt5.py"
    ```
"""

import asyncio
import logging
import time
from collections.abc import Callable, Coroutine
from functools import partial, wraps
from multiprocessing import Manager, Queue
from multiprocessing.managers import BaseManager
from threading import Thread
from typing import Any

from telegram import (
    CallbackQuery,
    InlineKeyboardButton,
    InlineKeyboardMarkup,
    KeyboardButton,
    ReplyKeyboardMarkup,
    Update,
)
from telegram.constants import MessageLimit, ParseMode
from telegram.error import BadRequest, NetworkError, TelegramError
from telegram.ext import (
    Application,
    CallbackContext,
    CallbackQueryHandler,
    CommandHandler,
)
from telegram.helpers import escape_markdown

from lettrade.plot import BotPlotter
from lettrade.stats import BotStatistic

from .commander import Commander

logger = logging.getLogger(__name__)


def authorized_only(command_handler: Callable[..., Coroutine[Any, Any, None]]):
    """Decorator to check if the message comes from the correct chat_id

    Args:
        command_handler (Callable[..., Coroutine[Any, Any, None]]): Telegram CommandHandler
    """

    @wraps(command_handler)
    async def wrapper(self: "TelegramCommander", *args, **kwargs):
        """Decorator logic"""
        update = kwargs.get("update") or args[0]

        # Reject unauthorized messages
        if update.callback_query:
            cchat_id = int(update.callback_query.message.chat.id)
        else:
            cchat_id = int(update.message.chat_id)

        if cchat_id != self._chat_id:
            logger.info(f"Rejected unauthorized message from: {update.message.chat_id}")
            return wrapper

        logger.debug(
            "Executing handler: %s for chat_id: %s",
            command_handler.__name__,
            self._chat_id,
        )
        try:
            return await command_handler(self, *args, **kwargs)
        except Exception as e:
            await self._send_msg(str(e))
            # except BaseException:
            logger.exception("Exception occurred within Telegram module", exc_info=e)

    return wrapper


class TelegramAPI:
    """Singleton object communicate across multipprocessing"""

    _app: Application
    _loop: asyncio.AbstractEventLoop
    _bots_queue: dict[str, Queue]
    _bot_selected: str

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_singleton"):
            cls._singleton = object.__new__(cls)
            cls._singleton.__init__(*args, **kwargs)
        return cls._singleton

    def __init__(self, token: str, chat_id: int, *args, **kwargs) -> None:
        self._token: str = token
        self._chat_id: int = int(chat_id)
        self._bots_queue = dict()
        self._bot_selected = None

    def start(self, pname: str, action_queue: Queue):
        """Start"""
        if pname in self._bots_queue:
            logger.warning("Process name %s override existed action queue", pname)
        self._bots_queue[pname] = action_queue

        logger.info("New join process: %s", pname)

        # TODO: Lock for safe multipleprocessing
        if hasattr(self, "_keyboard"):
            return

        self._init_keyboard()
        self._start_thread()

    def _action(self, action: str, pname: str | None = None):
        if pname is None:
            pname = list(self._bots_queue.keys())
        elif isinstance(pname, str):
            pname = [pname]

        for name in pname:
            q: Queue = self._bots_queue[name]
            q.put(action)

    def send_message(self, msg: str, pname: str, **kwargs) -> None:
        """Send message to Telegram Bot

        Args:
            msg (str): Message

        Returns:
            _type_: `None`
        """
        msg = f"*[Process: {pname}]*\n\n{escape_markdown(msg)}"
        asyncio.run_coroutine_threadsafe(self._send_msg(msg, **kwargs), self._loop)

    async def _cleanup_telegram(self) -> None:
        if self._app.updater:
            await self._app.updater.stop()
        await self._app.stop()
        await self._app.shutdown()

    def cleanup(self) -> None:
        """Stops all running telegram threads."""
        # This can take up to `timeout` from the call to `start_polling`.
        asyncio.run_coroutine_threadsafe(self._cleanup_telegram(), self._loop)
        self._thread.join()

    def _start_thread(self):
        """
        Creates and starts the polling thread
        """
        self._thread = Thread(target=self._init, name="TelegramAPI")
        self._thread.start()

    def _init_keyboard(self) -> None:
        """
        Validates the keyboard configuration from telegram config
        section.
        """
        self._keyboard: list[list[str | KeyboardButton]] = [
            ["/help", "/bots", "/bot"],
            ["/stats", "/plot", "/status"],
            ["/count"],
        ]

    def _init_telegram_app(self):
        return Application.builder().token(self._token).connection_pool_size(50).build()

    def _init(self) -> None:
        """
        Initializes this module with the given config,
        registers all known command handlers
        and starts polling for message updates
        Runs in a separate thread.
        """
        try:
            self._loop = asyncio.get_running_loop()
        except RuntimeError:
            self._loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._loop)

        self._app = self._init_telegram_app()

        # Register command handler and start telegram message polling
        handles = [
            CommandHandler("help", self._cmd_help),
            CommandHandler("bots", self._cmd_bots),
            CommandHandler("bot", self._cmd_bot),
            CommandHandler("status", self._cmd_status),
            CommandHandler("stats", self._cmd_stats),
            CommandHandler("plot", self._cmd_plot),
            # CommandHandler("balance", self._balance),
            # CommandHandler("start", self._start),
            # CommandHandler("stop", self._stop),
            # CommandHandler(["forcesell", "forceexit", "fx"], self._force_exit),
            # CommandHandler("reload_trade", self._reload_trade_from_exchange),
            # CommandHandler("trades", self._trades),
            # CommandHandler("delete", self._delete_trade),
            # CommandHandler(["coo", "cancel_open_order"], self._cancel_open_order),
            # CommandHandler("performance", self._performance),
            # CommandHandler(["buys", "entries"], self._enter_tag_performance),
            # CommandHandler(["sells", "exits"], self._exit_reason_performance),
            # CommandHandler("mix_tags", self._mix_tag_performance),
            # CommandHandler("daily", self._daily),
            # CommandHandler("weekly", self._weekly),
            # CommandHandler("monthly", self._monthly),
            # CommandHandler("count", self._count),
            # CommandHandler("locks", self._locks),
            # CommandHandler(["unlock", "delete_locks"], self._delete_locks),
            # CommandHandler(["reload_config", "reload_conf"], self._reload_config),
            # CommandHandler(["show_config", "show_conf"], self._show_config),
            # CommandHandler(["stopbuy", "stopentry"], self._stopentry),
            # CommandHandler("whitelist", self._whitelist),
            # CommandHandler("blacklist", self._blacklist),
            # CommandHandler(["blacklist_delete", "bl_delete"], self._blacklist_delete),
            # CommandHandler("logs", self._logs),
            # CommandHandler("edge", self._edge),
            # CommandHandler("health", self._health),
            # CommandHandler("version", self._version),
            # CommandHandler("marketdir", self._changemarketdir),
            # CommandHandler("order", self._order),
            # CommandHandler("list_custom_data", self._list_custom_data),
        ]
        callbacks = [
            # CallbackQueryHandler(self._status_table, pattern="update_status_table"),
            # CallbackQueryHandler(self._daily, pattern="update_daily"),
            # CallbackQueryHandler(self._weekly, pattern="update_weekly"),
            # CallbackQueryHandler(self._monthly, pattern="update_monthly"),
            # CallbackQueryHandler(self._profit, pattern="update_profit"),
            # CallbackQueryHandler(self._balance, pattern="update_balance"),
            # CallbackQueryHandler(self._performance, pattern="update_performance"),
            # CallbackQueryHandler(
            #     self._enter_tag_performance, pattern="update_enter_tag_performance"
            # ),
            # CallbackQueryHandler(
            #     self._exit_reason_performance, pattern="update_exit_reason_performance"
            # ),
            # CallbackQueryHandler(
            #     self._mix_tag_performance, pattern="update_mix_tag_performance"
            # ),
            # CallbackQueryHandler(self._count, pattern="update_count"),
            # CallbackQueryHandler(self._force_exit_inline, pattern=r"force_exit__\S+"),
            CallbackQueryHandler(self._run_cmd, pattern=r"cmd:\S+"),
        ]
        for handle in handles:
            self._app.add_handler(handle)

        for callback in callbacks:
            self._app.add_handler(callback)

        logger.info(
            "Telegram commander is listening for following commands: %s",
            [[x for x in sorted(h.commands)] for h in handles],
        )
        self._loop.run_until_complete(self._startup_telegram())

    async def _startup_telegram(self) -> None:
        await self._app.initialize()
        await self._app.start()
        if self._app.updater:
            await self._app.updater.start_polling(
                bootstrap_retries=-1,
                timeout=20,
                # read_latency=60,  # Assumed transmission latency
                drop_pending_updates=True,
                # stop_signals=[],  # Necessary as we don't run on the main thread
            )
            while True:
                await asyncio.sleep(10)
                if not self._app.updater.running:
                    break

    async def _send_msg(
        self,
        msg: str,
        parse_mode: str = ParseMode.MARKDOWN,
        disable_notification: bool = False,
        keyboard: list[list[InlineKeyboardButton]] | None = None,
        callback_path: str = "",
        reload_able: bool = False,
        query: CallbackQuery | None = None,
    ) -> None:
        """
        Send given markdown message
        :param msg: message
        :param bot: alternative bot
        :param parse_mode: telegram parse mode
        :return: None
        """
        reply_markup: InlineKeyboardMarkup | ReplyKeyboardMarkup
        # if query:
        #     await self._update_msg(
        #         query=query,
        #         msg=msg,
        #         parse_mode=parse_mode,
        #         callback_path=callback_path,
        #         reload_able=reload_able,
        #     )
        #     return
        # if reload_able and self._config["telegram"].get("reload", True):
        #     reply_markup = InlineKeyboardMarkup(
        #         [[InlineKeyboardButton("Refresh", callback_data=callback_path)]]
        #     )
        # else:
        if keyboard is not None:
            reply_markup = InlineKeyboardMarkup(keyboard)
        else:
            reply_markup = ReplyKeyboardMarkup(self._keyboard, resize_keyboard=True)
        try:
            try:
                await self._app.bot.send_message(
                    self._chat_id,
                    text=msg,
                    parse_mode=parse_mode,
                    reply_markup=reply_markup,
                    disable_notification=disable_notification,
                )
            except NetworkError as network_err:
                # Sometimes the telegram server resets the current connection,
                # if this is the case we send the message again.
                logger.warning(
                    "Telegram NetworkError: %s! Trying one more time.",
                    network_err.message,
                )
                await self._app.bot.send_message(
                    self._chat_id,
                    text=msg,
                    parse_mode=parse_mode,
                    reply_markup=reply_markup,
                    disable_notification=disable_notification,
                )
        except TelegramError as telegram_err:
            logger.warning(
                "TelegramError: %s! Giving up on that message.", telegram_err.message
            )

    ##### COMMAND #####
    @authorized_only
    async def _cmd_help(self, update: Update, context: CallbackContext) -> None:
        """Handler for /help.
        Show commands of the bot

        Args:
            update (Update): message update
            context (CallbackContext): Telegram context
        """
        msg = (
            "_Bots Control_\n"
            "*/bots:* `Show list of bot`\n"
            "*/bot <bot_index>:* `Select a bot`\n"
            "\n_Bot Control_\n"
            "------------\n"
            "*/balance:* `Show bot managed balance per currency`\n"
            "*/logs [limit]:* `Show latest logs - defaults to 10` \n"
            "*/count:* `Show number of active trades compared to allowed number of trades`\n"
            "*/health* `Show latest process timestamp - defaults to 1970-01-01 00:00:00` \n"
            "\n_Statistics_\n"
            "------------\n"
            "*/status <trade_id>|[table]:* `Lists all open trades`\n"
            "*/trades [limit]:* `Lists last closed trades (limited to 10 by default)`\n"
            "*/profit [<n>]:* `Lists cumulative profit from all finished trades`\n"
            "*/stats:* `Shows Wins / losses by Sell reason as well as "
            "Avg. holding durations for buys and sells.`\n"
            "*/help:* `This help message`\n"
            "*/version:* `Show version`\n"
        )

        await self._send_msg(msg, parse_mode=ParseMode.MARKDOWN)

    @authorized_only
    async def _cmd_status(self, update: Update, context: CallbackContext) -> None:
        """Handler for /status.
        Returns the current Trade status

        Args:
            update (Update): message update
            context (CallbackContext): Telegram context
        """
        lines = ["Hello", "LetTrade"]
        r = dict()

        msg = ""
        for line in lines:
            if line:
                # if (len(msg) + len(line) + 1) < MAX_MESSAGE_LENGTH:
                msg += line + "\n"
                # # else:
                # await self._send_msg(msg.format(**r))
                # msg = "*Trade ID:* `{trade_id}` - continued\n" + line + "\n"

        await self._send_msg(msg.format(**r))

    @authorized_only
    async def _cmd_bots(self, update: Update, context: CallbackContext) -> None:
        """Handler for /bots
        Returns the current Strategy Statistic

        Args:
            update (Update): message update
            context (CallbackContext): Telegram context
        """
        msg = "_Listening Bots_"
        keyboard = []
        for i, name in enumerate(list(self._bots_queue.keys())):
            msg += f"\n*{i}*: [*{escape_markdown(name,2)}*](/bot {i})"
            keyboard.append(
                [
                    InlineKeyboardButton(
                        text=f"/bot {i} {name}",
                        callback_data=f"cmd:/bot {i}",
                    )
                ]
            )

        # print("Telegram: bots:\n", msg, keyboard)
        await self._send_msg(msg, parse_mode=ParseMode.MARKDOWN_V2, keyboard=keyboard)

    @authorized_only
    async def _cmd_bot(self, update: Update, context: CallbackContext) -> None:
        """Handler for /bot
        Returns the current Strategy Statistic

        Args:
            update (Update): message update
            context (CallbackContext): Telegram context
        """
        if (
            not context.args
            or not len(context.args) >= 1
            or not context.args[0].isnumeric()
        ):
            msg = (
                "*/bot <bot_index>*: to select a bot"
                "\n*/bots*: to get list of bot index"
            )
        else:
            bot_index = int(context.args[0])
            names = list(self._bots_queue.keys())

            if bot_index < 0 or bot_index >= len(names):
                msg = f"_Error select bot: {bot_index} "
                msg += f"is out of range {len(names)}_"
            else:
                bot_name = names[bot_index]
                self._bot_selected = bot_name
                msg = "_Success select bot_"

        if self._bot_selected is not None:
            msg += f"\n------\nSelected bot: *{self._bot_selected}*"
            keyboard = [
                [
                    InlineKeyboardButton(
                        text=f"/plot {self._bot_selected}",
                        callback_data=f"cmd:/plot {self._bot_selected}",
                    )
                ]
            ]
        else:
            keyboard = None

        await self._send_msg(msg, keyboard=keyboard)

    @authorized_only
    async def _cmd_stats(self, update: Update, context: CallbackContext) -> None:
        """Handler for /stats
        Returns the current Strategy Statistic

        Args:
            update (Update): message update
            context (CallbackContext): Telegram context
        """
        # stats: Statistic = self.bot.stats
        # stats.compute()
        # await self._send_msg(stats.result.to_string())

        self._action("stats")

    def on_stats(self, stats: str, pname: str):
        self.send_message(stats, pname=pname)

    @authorized_only
    async def _cmd_plot(self, update: Update, context: CallbackContext) -> None:
        """Handler for /plot
        Returns the current Strategy Statistic

        Args:
            update (Update): message update
            context (CallbackContext): Telegram context
        """
        # plot: Statistic = self.bot.plot
        # plot.compute()
        # await self._send_msg(plot.result.to_string())

        self._action("plot", pname=self._bot_selected)

    def on_plot(self, plot: str, pname: str):
        self.send_message(plot, pname=pname)

    @authorized_only
    async def _run_cmd(self, update: Update, context: CallbackContext) -> None:
        if not update.callback_query:
            return

        query = update.callback_query
        if not query or not query.data or not query.data.startswith("cmd:"):
            return

        datas = query.data.split("cmd:", 1)[1].split(" ")

        logger.info("Run command: %s", datas)

        context.args = datas[1:]
        match datas[0]:
            case "/bot":
                await self._cmd_bot(update=update, context=context)
            case "/plot":
                await self._cmd_plot(update=update, context=context)


class TelegramCommander(Commander):
    """Send notify and receive command from Telegram Bot"""

    _api: TelegramAPI
    _is_running: bool

    @classmethod
    def multiprocess(cls, kwargs, **other_kwargs):
        BaseManager.register("TelegramAPI", TelegramAPI)
        manager = BaseManager()
        manager.start()
        kwargs["api"] = manager.TelegramAPI(**kwargs)

    def __init__(
        self,
        token: str,
        chat_id: int,
        api: TelegramAPI | None = None,
        *args,
        **kwargs,
    ) -> None:
        """_summary_

        Args:
            token (str): Telegram Bot token
            chat_id (int): Telegram chat_id
        """
        super().__init__(*args, **kwargs)
        self._api = api or TelegramAPI(token=token, chat_id=chat_id)
        self._is_running = True

    def start(self):
        """Start"""
        logger.info("TelegramCommander start %s", self._name)
        q = self._t_action()
        self._api.start(pname=self._name, action_queue=q)

    def stop(self):
        """Stop"""
        logger.info("TelegramCommander stop %s", self._name)
        self._api.cleanup()
        self._is_running = False

    def send_message(self, msg: str, **kwargs) -> None:
        self._api.send_message(msg=msg, pname=self._name, **kwargs)

    def _t_action(self) -> Queue:
        manager = Manager()
        q = manager.Queue(maxsize=1_000)
        t = Thread(target=self._on_action, kwargs=dict(q=q))
        t.start()
        return q

    def _on_action(self, q: Queue):
        while self._is_running:
            try:
                action = q.get()

                match action:
                    case "stats":
                        self._on_action_stats()
                    case "plot":
                        self._on_action_plot()
            except (BrokenPipeError, EOFError) as e:
                # logger.error("Action", exc_info=e)
                self._is_running = False
                raise e
            except Exception as e:
                logger.error("Action", exc_info=e)
                time.sleep(5)

    def _on_action_stats(self):
        stats: BotStatistic = self.bot.stats
        stats.compute()
        self._api.on_stats(stats=stats.result.to_string(), pname=self._name)

    def _on_action_plot(self):
        plotter: BotPlotter = self.bot.plotter
        plotter.load()
        plotter.plot()
        # self._api.on_plot(plot=plot.result.to_string(), pname=self._name)



from .candlestick import *
from .dataframe import *
from .momentum import *
from .plot import DATAFRAME_PLOTTERS_NAME, IndicatorPlotter, indicator_load_plotters
from .series import *
from .trend import *
from .volatility import *


def indicators_inject_pandas(obj: object | None = None):
    """Inject indicators to Pandas"""
    if obj is None:
        from pandas.core.base import PandasObject

        obj = PandasObject

    # Flag to mark indicators injected
    if hasattr(obj, "_lt_indicators_injected"):
        return

    from .candlestick import pandas_inject as candlestick_pandas_inject
    from .dataframe import pandas_inject as dataframe_pandas_inject
    from .momentum import pandas_inject as momentum_pandas_inject
    from .series import pandas_inject as series_pandas_inject
    from .trend import pandas_inject as trend_pandas_inject
    from .volatility import pandas_inject as volatility_pandas_inject

    series_pandas_inject(obj)
    dataframe_pandas_inject(obj)

    candlestick_pandas_inject(obj)
    trend_pandas_inject(obj)
    volatility_pandas_inject(obj)
    momentum_pandas_inject(obj)

    # Flag to mark indicators injected
    obj._lt_indicators_injected = True



import logging
import queue
import threading
import time
from multiprocessing import Manager, Queue
from multiprocessing.managers import SyncManager

import pandas as pd

from .plot import OptimizePlotter

logger = logging.getLogger(__name__)


class OptimizeStatistic:
    """
    Compute strategy result
    """

    _q: Queue
    _total: int = 0
    _manager: SyncManager
    _result_thread: threading.Thread | None

    results: list
    result: pd.Series
    plotter: OptimizePlotter = None

    def __init__(self, plotter: OptimizePlotter = None, total: int = 0) -> None:
        self.plotter = plotter
        self._total = total

        self.results = []
        self._result_thread = None

        if self.plotter is not None:
            self.plotter.init(self.results)

        self._t_wait_result()

    @property
    def queue(self) -> Queue:
        return self._q

    def _t_wait_result(self):
        self._manager = Manager()
        self._q = self._manager.Queue(maxsize=1_000)
        self._result_thread = threading.Thread(target=self._wait_done)
        self._result_thread.start()

    def _wait_done(self, retry=10):
        done = 0
        while True:
            try:
                result = self._q.get(timeout=3)
            except queue.Empty:
                continue
            except (AttributeError, BrokenPipeError):
                break
            except Exception as e:
                # TODO: check closed main class then exit
                retry -= 1
                if retry <= 0:
                    logger.warning("queue %s", e, exc_info=e)
                    break
                continue

            if result is None:
                return

            self.results.append(result)

            done += 1

            if self.plotter:
                self.plotter.on_result(result)

            # Done
            if self._total > 0 and done == self._total:
                break

        # self.done()

    def done(self):
        if self._result_thread is not None:
            self._q.put(None)
            self._result_thread.join()
            self._result_thread = None

        if self.plotter:
            self.plotter.on_done()

        time.sleep(1)  # Wait for return finish

        if self._manager is not None:
            self._manager.shutdown()
            del self._manager
            self._manager = None

        if self._q is not None:
            try:
                self._q.close()
                del self._q
            except Exception:
                pass
            self._q = None

    def compute(self):
        """
        Calculate strategy report
        """

        return self.result

    def stop(self):
        self.done()



import logging
from concurrent.futures import ProcessPoolExecutor
from typing import TYPE_CHECKING

from lettrade.account import Account
from lettrade.bot import LetTradeBot
from lettrade.commander import Commander
from lettrade.data import DataFeed, DataFeeder
from lettrade.exchange import Exchange
from lettrade.plot import BotPlotter, Plotter
from lettrade.stats import BotStatistic
from lettrade.strategy import Strategy

if TYPE_CHECKING:
    from lettrade.exchange.backtest import OptimizePlotter

logger = logging.getLogger(__name__)


class LetTrade:
    """Building new bot object and handle multiprocessing"""

    _plotter: Plotter | None = None
    """Plot graphic results"""
    _stats: BotStatistic | None = None
    _bot: LetTradeBot | None = None

    _kwargs: dict

    def __init__(
        self,
        datas: DataFeed | list[DataFeed] | str | list[str],
        strategy: type[Strategy],
        feeder: type[DataFeeder],
        exchange: type[Exchange],
        account: type[Account],
        commander: type[Commander] | Commander | None = None,
        stats: type[BotStatistic] = BotStatistic,
        plotter: type[Plotter] | None = None,
        name: str | None = None,
        bot: type[LetTradeBot] = LetTradeBot,
        **kwargs,
    ) -> None:
        """_summary_

        Args:
            datas (DataFeed | list[DataFeed] | str | list[str]): _description_
            strategy (type[Strategy]): _description_
            feeder (type[DataFeeder]): _description_
            exchange (type[Exchange]): _description_
            account (type[Account]): _description_
            commander (type[Commander] | Commander | None, optional): _description_. Defaults to None.
            stats (type[BotStatistic], optional): _description_. Defaults to BotStatistic.
            plotter (type[Plotter] | None, optional): _description_. Defaults to None.
            name (str | None, optional): _description_. Defaults to None.
            bot (type[LetTradeBot], optional): _description_. Defaults to LetTradeBot.
        """
        self._kwargs = kwargs
        self._kwargs["strategy_cls"] = strategy
        self._kwargs["feeder_cls"] = feeder
        self._kwargs["exchange_cls"] = exchange
        self._kwargs["account_cls"] = account
        self._kwargs["commander_cls"] = commander
        self._kwargs["stats_cls"] = stats
        self._kwargs["plotter_cls"] = plotter
        self._kwargs["bot_cls"] = bot
        self._kwargs["name"] = name

        self._kwargs["datas"] = self._init_datafeeds(datas)

    def _datafeed(self, data: DataFeed, **kwargs) -> DataFeed:
        """Init and validate DataFeed

        Args:
            data (DataFeed): DataFeed or config of it

        Raises:
            RuntimeError: Validate error

        Returns:
            DataFeed: The DataFeed
        """
        match data:
            case DataFeed():
                return data
            case _:
                raise RuntimeError(f"data {data} is invalid")

    def _init_datafeeds(self, datas) -> list[DataFeed] | list[list[DataFeed]]:
        # Support single and multiple data
        if not isinstance(datas, list):
            datas = [datas]

        # Check data
        feeds: list[list[DataFeed]] | list[DataFeed]
        if isinstance(datas[0], list):
            feeds = []
            for i, data in enumerate(datas):
                data_feeds: list[DataFeed] = []
                for j, d in enumerate(data):
                    df = self._datafeed(data=d, index=i * j)
                    data_feeds.append(df)
                feeds.append(data_feeds)
        else:
            feeds = [self._datafeed(data=data, index=i) for i, data in enumerate(datas)]
        return feeds

    def start(self, force: bool = False):
        """Start LetTrade by init bot object and loading datafeeds

        Args:
            force (bool, optional): _description_. Defaults to False.
        """
        if force and self._bot is not None:
            self._bot = None

        self._bot = self._bot_cls.start_bot(bot=self._bot, **self._kwargs)

    def run(self, worker: int | None = None, **kwargs):
        """Run LetTrade in single or multiple processing

        Args:
            worker (int | None, optional): Number of processing. Defaults to None.
        """
        if worker or isinstance(self.data, list):
            if not isinstance(self.data, list):
                # self.data = self.datas
                self.datas = [self.data]

            worker = self._max_workers(worker)

            self._multiprocess()

            datas_source = self._kwargs.pop("datas")
            with ProcessPoolExecutor(max_workers=worker) as executor:
                futures = [
                    executor.submit(
                        self._bot_cls.run_bot,
                        datas=datas,
                        id=i,
                        result="str",
                        **self._kwargs,
                        **kwargs,
                    )
                    for i, datas in enumerate(datas_source)
                ]
                for future in futures:
                    result = future.result()
                    print(result)
        else:
            self._bot = self._bot_cls.run_bot(
                bot=self._bot,
                result="bot",
                **self._kwargs,
                **kwargs,
            )
            print(str(self._bot.stats))

    def _max_workers(self, worker):
        if not worker:
            worker = len(self.datas)
        elif worker > len(self.datas):
            logger.warning(
                "Worker size %s is more then datas size %s",
                worker,
                len(self.datas),
            )
            worker = len(self.datas)
        return worker

    def _multiprocess(self, **kwargs):
        if self._bot is not None:
            logger.warning("Removed exist bot when running in multiprocessing")
            self._bot = None

        if self._commander_cls:
            # Implement commander dependencies and save to commander_kwargs
            commander_kwargs = self._kwargs.setdefault("commander_kwargs", {})
            self._commander_cls.multiprocess(
                kwargs=commander_kwargs,
                # self_kwargs=self._kwargs,
                **kwargs,
            )

    def stop(self):
        """Stop LetTrade"""
        if self._bot is not None:
            return self._bot.stop()
        if self.stats:
            self.stats.stop()

    @property
    def stats(self) -> BotStatistic | None:
        if self._stats:
            return self._stats
        if self._bot is not None:
            return self._bot.stats
        return None

    @property
    def commander(self) -> Commander | None:
        if self._bot is not None:
            return self._bot.commander
        return None

    # Plotter
    @property
    def plotter(self) -> "Plotter | BotPlotter | OptimizePlotter | None":
        if self._plotter is not None:
            return self._plotter
        if self._bot is not None:
            return self._bot.plotter
        raise RuntimeError("Plotter is not defined")

    def plot(self, *args, jump: dict | None = None, **kwargs):
        """Plot strategy/optimize result

        BotPlotter:
            Miror of [BotPlotter.plot()](site:/reference/plot/bot/#lettrade.plot.bot.BotPlotter.plot).
            Plotly implement [PlotlyBotPlotter.plot()](site:/reference/plot/plotly/plotly/#lettrade.plot.plotly.plotly.PlotlyBotPlotter.plot).

            Args:
                `jump` (dict | None, optional): Miror of [BotPlotter.jump()](site:/reference/plot/bot/#lettrade.plot.bot.BotPlotter.jump)

            Example:
                Jump to position_id
                    ```python
                    lt.plot(
                        jump=dict(position_id=1, range=300),
                        layout=dict(height=2000),
                    )
                    ```

        OptimizePlotter:
            Miror of [OptimizePlotter.plot()](site:/reference/plot/optimize/#lettrade.plot.optimize.OptimizePlotter.plot).
            Plotly implement [PlotlyOptimizePlotter.plot()](site:/reference/exchange/backtest/plotly/optimize/#lettrade.exchange.backtest.plotly.optimize.PlotlyOptimizePlotter.plot).

            Example:
                    ```python
                    lt.plot(layout=dict(height=2000))
                    ```
        """
        if __debug__:
            from .utils.docs import is_docs_session

            if is_docs_session():
                return

        self.plotter.plot(*args, jump=jump, **kwargs)

    # Bot kwargs properties
    @property
    def name(self) -> type[LetTradeBot]:
        return self._kwargs.get("name", None)

    @property
    def datas(self) -> list[DataFeed]:
        return self._kwargs.get("datas", None)

    @datas.setter
    def datas(self, value: list[DataFeed]) -> None:
        self._kwargs["datas"] = value

    @property
    def data(self) -> DataFeed:
        return self.datas[0]

    @property
    def _bot_cls(self) -> type[LetTradeBot]:
        return self._kwargs.get("bot_cls", None)

    @property
    def _commander_cls(self) -> type[Commander]:
        return self._kwargs.get("commander_cls", None)

    @_commander_cls.setter
    def _commander_cls(self, value):
        self._kwargs["commander_cls"] = value

    @property
    def _stats_cls(self) -> type[BotStatistic]:
        return self._kwargs.get("stats_cls", None)

    @property
    def _plotter_cls(self) -> type["Plotter"]:
        return self._kwargs.get("plotter_cls", None)

    @_plotter_cls.setter
    def _plotter_cls(self, value):
        self._kwargs["plotter_cls"] = value



import logging
from typing import TYPE_CHECKING

from .base import BaseTransaction

if TYPE_CHECKING:
    from lettrade.data import DataFeed

    from .exchange import Exchange
    from .order import Order
    from .position import Position

logger = logging.getLogger(__name__)


class Execution(BaseTransaction):
    """Execution"""

    def __init__(
        self,
        id: str,
        exchange: "Exchange",
        data: "DataFeed",
        size: float,
        price: float,
        at: float,
        order_id: str | None = None,
        order: "Order | None" = None,
        position_id: str | None = None,
        position: "Position | None" = None,
        **kwargs,
    ):
        super().__init__(
            id=id,
            exchange=exchange,
            data=data,
            size=size,
            **kwargs,
        )
        self.order_id = order_id
        self.order: "Order" = order
        self.position_id = position_id
        self.position: "Position" = position
        self.price = price
        self.at = at

    def __repr__(self):
        return f"<{self.__class__.__name__} id={self.id} size={self.size}>"

    def _on_execution(self):
        self.exchange.on_execution(self)

    def merge(self, other: "Execution"):
        """
        Merge to keep object handler but not overwrite for Strategy using when Strategy want to store object and object will be automatic update directly
        """
        if other is self:
            return

        if self.id != other.id:
            raise RuntimeError(f"Merge difference id {self.id} != {other.id} execution")

        self.price = other.price
        self.size = other.size

        if other.at:
            self.at = other.at
        if other.order_id:
            self.order_id = other.order_id
        if other.order:
            self.order = other.order
        if other.position_id:
            self.position_id = other.position_id
        if other.position:
            self.position = other.position



import plotly.io as pio

from lettrade.plot.helper import plot_merge

from .indicator import *
from .plotly import *

pio.templates.default = "plotly_dark"

if __debug__:  # __debug__ to remove code in production
    from lettrade.utils.notebook import is_notebook_session

    if is_notebook_session():
        pio.renderers.default = "notebook"



from .account import BackTestAccount, ForexBackTestAccount
from .backtest import LetTradeBackTest, LetTradeBackTestBot, let_backtest
from .commander import BackTestCommander, StorageBackTestCommander
from .data import BackTestDataFeed, CSVBackTestDataFeed, YFBackTestDataFeed
from .exchange import BackTestExchange
from .feeder import BackTestDataFeeder
from .plot import OptimizePlotter
from .trade import BackTestExecution, BackTestOrder, BackTestPosition



if __debug__:
    validate_strategy_trade: bool = False
    """Flag to validate trading object is not writing attribute from strategy"""
    validate_data_getitem_pointer: bool = False
    """Flag to validate `DataFeed` get data by pointer but doesn't has prefix `DataFeed.l.`"""



from lettrade.base.error import LetException
from lettrade.exchange import LetOrderInvalidException


class LetLiveException(LetException):
    raw: dict

    def __init__(self, *args: object, raw: dict = None) -> None:
        super().__init__(*args)
        self.raw = raw


class LetLiveAPIUnauthorizedException(LetLiveException):
    """API Unauthorized exception"""


class LetLiveOrderInvalidException(LetLiveException, LetOrderInvalidException):
    """Live order place is invalid exception"""



from .candlestick import cdl_direction
from .talib import (
    cdl_3blackcrows,
    cdl_3inside,
    cdl_3whitesoldiers,
    cdl_doji,
    cdl_engulfing,
    cdl_eveningstar,
    cdl_morningstar,
    cdl_pattern,
)


def pandas_inject(obj: object | None = None):
    if obj is None:
        from pandas.core.base import PandasObject

        obj = PandasObject

    obj.cdl_direction = cdl_direction

    obj.cdl_pattern = cdl_pattern
    obj.cdl_3blackcrows = cdl_3blackcrows
    obj.cdl_3whitesoldiers = cdl_3whitesoldiers
    obj.cdl_morningstar = cdl_morningstar
    obj.cdl_eveningstar = cdl_eveningstar
    obj.cdl_doji = cdl_doji
    obj.cdl_3inside = cdl_3inside
    obj.cdl_engulfing = cdl_engulfing



import logging

from lettrade.data import DataFeed
from lettrade.exchange import Exchange, OrderResult, OrderType, PositionState

from .api import LiveAPI
from .trade import LiveExecution, LiveOrder, LivePosition

logger = logging.getLogger(__name__)


class LiveExchange(Exchange):
    """MetaTrade 5 exchange module for `lettrade`"""

    _execution_cls: type[LiveExecution] = LiveExecution
    _order_cls: type[LiveOrder] = LiveOrder
    _position_cls: type[LivePosition] = LivePosition

    _api: LiveAPI

    def __init__(self, api: LiveAPI, *args, **kwargs):
        """_summary_

        Args:
            api (LiveAPI): API connect to rpyc MeTrader 5 Terminal server through module `mt5linux`
            *args (list): `Exchange` list parameters
            **kwargs (dict): `Exchange` dict parameters
        """
        super().__init__(*args, **kwargs)
        self._api = api

    def start(self) -> None:
        """Start Live exchange by: Sync positions from server"""
        self._api.start(exchange=self)
        return super().start()

    def next(self) -> None:
        self._api.next()
        return super().next()

    def new_order(
        self,
        size: float,
        type: OrderType | None = OrderType.Market,
        limit: float | None = None,
        stop: float | None = None,
        sl: float | None = None,
        tp: float | None = None,
        tag: str | None = None,
        data: DataFeed | None = None,
        **kwargs,
    ) -> OrderResult:
        """Place new order to server

        Args:
            size (float): _description_
            type (OrderType | None, optional): _description_. Defaults to OrderType.Market.
            limit (float | None, optional): _description_. Defaults to None.
            stop (float | None, optional): _description_. Defaults to None.
            sl (float | None, optional): _description_. Defaults to None.
            tp (float | None, optional): _description_. Defaults to None.
            tag (str | None, optional): _description_. Defaults to None.
            data (DataFeed | None, optional): _description_. Defaults to None.

        Returns:
            OrderResult: _description_
        """
        if not data:
            data = self.data

        if type == OrderType.Market:
            limit = 0
            stop = 0

        order = self._order_cls(
            id=-1,
            exchange=self,
            data=data,
            size=size,
            type=type,
            limit_price=limit,
            stop_price=stop,
            sl_price=sl,
            tp_price=tp,
            tag=tag,
            **kwargs,
        )
        ok = order.place()

        # if __debug__:
        #     logger.info("New order %s at %s", order, self.data.now)

        return ok

    # Events
    def on_executions_event(self, raws, broadcast: bool | None = True):
        if __debug__:
            logger.debug("Raw new executions: %s", raws)
        for raw in raws:
            execution = self._execution_cls.from_raw(raw=raw, exchange=self)
            if execution is None:
                continue
            self.on_execution(execution, broadcast=broadcast)

    def on_orders_event(self, new: list = None, old: list = None):
        if __debug__:
            logger.debug("Raw orders new: %s, old: %s", new, old)

        if new is None:
            new = []
        if old is None:
            old = []

        orders = []
        for raw in [*new, *old]:
            order = self._order_cls.from_raw(raw=raw, exchange=self)
            if order is None:
                continue
            orders.append(order)

        if orders:
            self.on_orders(orders)

    def on_positions_event(self, new: list = None, old: list = None):
        if __debug__:
            logger.debug("Raw positions new: %s, old: %s", new, old)

        # Positions new
        positions_new = []
        if new:
            for raw in new:
                position = self._position_cls.from_raw(
                    raw=raw,
                    exchange=self,
                    state=PositionState.Open,
                )
                if position is None:
                    continue
                positions_new.append(position)

        # positions_old
        positions_old = []
        if old:
            for raw in old:
                position = self._position_cls.from_raw(
                    raw=raw,
                    exchange=self,
                    state=PositionState.Exit,
                )
                if position is None:
                    continue
                positions_old.append(position)

        # Event
        positions = [*positions_new, *positions_old]
        if positions:
            self.on_positions(positions)



import plotly.io as pio

from .optimize import PlotlyOptimizePlotter

pio.templates.default = "plotly_dark"

if __debug__:  # __debug__ to remove code in production
    from lettrade.utils.notebook import is_notebook_session

    if is_notebook_session():
        pio.renderers.default = "notebook"



import logging

from lettrade.data import DataFeed
from lettrade.exchange import Exchange, OrderResult, OrderType

from .trade import BackTestOrder

logger = logging.getLogger(__name__)


class BackTestExchange(Exchange):
    __id = 0

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if self._config.setdefault("use_execution", False):
            self.executions = None

    def _id(self) -> str:
        self.__id += 1
        return str(self.__id)

    def next(self):
        """Execution when new data feeded"""
        self._simulate_orders()
        super().next()

    def on_execution(self, *args, **kwargs) -> None:
        if self.executions is None:
            logger.warning(
                "Execution transaction is disable, enable by flag: show_execution=True"
            )
            return

        super().on_execution(*args, **kwargs)

    def new_order(
        self,
        size: float,
        type: OrderType = OrderType.Market,
        limit: float | None = None,
        stop: float | None = None,
        sl: float | None = None,
        tp: float | None = None,
        tag: object | None = None,
        data: DataFeed | None = None,
        **kwargs,
    ) -> OrderResult:
        """Place new order.
        Then send order events to `Brain`

        Args:
            size (float): _description_
            type (OrderType | None, optional): _description_. Defaults to OrderType.Market.
            limit (float | None, optional): _description_. Defaults to None.
            stop (float | None, optional): _description_. Defaults to None.
            sl (float | None, optional): _description_. Defaults to None.
            tp (float | None, optional): _description_. Defaults to None.
            tag (object | None, optional): _description_. Defaults to None.
            data (DataFeed | None, optional): _description_. Defaults to None.
            **kwargs (dict | None, optional): Extra-parameters

        Returns:
            OrderResult: Result when place new `Order`
        """
        if not data:
            data = self.data

        order = BackTestOrder(
            id=self._id(),
            exchange=self,
            data=data,
            size=size,
            type=type,
            limit_price=limit,
            stop_price=stop,
            sl_price=sl,
            tp_price=tp,
            tag=tag,
        )
        ok = order.place(at=self.data.bar())

        if type == OrderType.Market:
            # Simulate market order will send event before return order result
            self._simulate_orders()

        return ok

    def _simulate_orders(self):
        for order in list(self.orders.values()):
            self._simulate_order(order)

    def _simulate_order(self, order: BackTestOrder):
        if not order.is_opening:
            return

        if order.type == OrderType.Market:
            order.fill(
                price=self.data.l.open[0],
                at=self.data.bar(),
            )
            return

        if order.type == OrderType.Limit:
            if order.is_long:
                # Buy Limit
                price = self.data.l.low[-1]
                if order.limit_price > price:
                    order.fill(price=order.limit_price, at=self.data.bar(-1))
                    return
            else:
                # Sell Limit
                price = self.data.l.high[-1]
                if order.limit_price < price:
                    order.fill(price=order.limit_price, at=self.data.bar(-1))
                    return
        elif order.type == OrderType.Stop:
            if order.is_long:
                # Buy Stop
                price = self.data.l.high[-1]
                if order.stop_price < price:
                    order.fill(price=order.stop_price, at=self.data.bar(-1))
                    return
            else:
                # Sell Stop
                price = self.data.l.low[-1]
                if order.stop_price > price:
                    order.fill(price=order.stop_price, at=self.data.bar(-1))
                    return



import logging
from typing import Literal

from lettrade import BotStatistic, Commander, Plotter, Strategy
from lettrade.exchange.live import (
    LetTradeLive,
    LetTradeLiveBot,
    LiveAccount,
    LiveDataFeed,
    LiveDataFeeder,
    LiveExchange,
    let_live,
)
from lettrade.exchange.live.api import LiveAPI

from .api import CCXTAPI
from .trade import CCXTExecution, CCXTOrder, CCXTPosition

logger = logging.getLogger(__name__)


class CCXTDataFeed(LiveDataFeed):
    """DataFeed for CCXT"""

    _api_cls: type[CCXTAPI] = CCXTAPI


class CCXTDataFeeder(LiveDataFeeder):
    """DataFeeder for CCXT"""

    _api_cls: type[CCXTAPI] = CCXTAPI
    _data_cls: type[CCXTDataFeed] = CCXTDataFeed


class CCXTAccount(LiveAccount):
    """Account for CCXT"""

    _currency: str

    def __init__(self, api: LiveAPI, currency: str = "USDT", **kwargs) -> None:
        super().__init__(api, **kwargs)
        self._currency = currency


class CCXTExchange(LiveExchange):
    """MetaTrade 5 exchange module for `lettrade`"""

    _execution_cls: type[CCXTExecution] = CCXTExecution
    _order_cls: type[CCXTOrder] = CCXTOrder
    _position_cls: type[CCXTPosition] = CCXTPosition


class LetTradeCCXTBot(LetTradeLiveBot):
    """LetTradeBot for CCXT"""


class LetTradeCCXT(LetTradeLive):
    """Help to maintain CCXT bots"""

    _data_cls: type[CCXTDataFeed] = CCXTDataFeed

    def __init__(
        self,
        feeder: type[CCXTDataFeeder] = CCXTDataFeeder,
        exchange: type[CCXTExchange] = CCXTExchange,
        account: type[CCXTAccount] = CCXTAccount,
        **kwargs,
    ) -> None:
        """_summary_

        Args:
            feeder (Type[CCXTDataFeeder], optional): _description_. Defaults to CCXTDataFeeder.
            exchange (Type[CCXTExchange], optional): _description_. Defaults to CCXTExchange.
            account (Type[CCXTAccount], optional): _description_. Defaults to CCXTAccount.
        """
        super().__init__(
            feeder=feeder,
            exchange=exchange,
            account=account,
            **kwargs,
        )


def let_ccxt(
    datas: set[set[str]],
    strategy: type[Strategy],
    *,
    ccxt_exchange: str,
    ccxt_key: str,
    ccxt_secret: str,
    ccxt_type: Literal["spot", "margin", "future"] = "spot",
    ccxt_verbose: bool = False,
    feeder: type[CCXTDataFeeder] = CCXTDataFeeder,
    exchange: type[CCXTExchange] = CCXTExchange,
    account: type[CCXTAccount] = CCXTAccount,
    commander: type[Commander] | None = None,
    stats: type[BotStatistic] | None = BotStatistic,
    plotter: type[Plotter] | None = None,
    bot: type[LetTradeCCXTBot] | None = LetTradeCCXTBot,
    lettrade: type[LetTradeCCXT] | None = LetTradeCCXT,
    api: type[CCXTAPI] | None = CCXTAPI,
    **kwargs,
) -> LetTradeCCXT:
    """Help to build `LetTradeCCXT`

    Args:
        datas (set[set[str]]): _description_
        strategy (Type[Strategy]): _description_
        ccxt_exchange (str): _description_
        ccxt_key (str): _description_
        ccxt_secret (str): _description_
        ccxt_type (Literal["spot", "margin", "future"], optional): _description_. Defaults to "spot".
        ccxt_verbose (bool, optional): _description_. Defaults to False.
        feeder (Type[CCXTDataFeeder], optional): _description_. Defaults to CCXTDataFeeder.
        exchange (Type[CCXTExchange], optional): _description_. Defaults to CCXTExchange.
        account (Type[CCXTAccount], optional): _description_. Defaults to CCXTAccount.
        commander (Type[Commander] | None, optional): _description_. Defaults to None.
        stats (Type[BotStatistic] | None, optional): _description_. Defaults to BotStatistic.
        plotter (Type[Plotter] | None, optional): _description_. Defaults to None.
        bot (Type[LetTradeCCXTBot] | None, optional): _description_. Defaults to LetTradeCCXTBot.
        lettrade (Type[LetTradeCCXT] | None, optional): _description_. Defaults to LetTradeCCXT.
        api (Type[CCXTAPI] | None, optional): _description_. Defaults to CCXTAPI.
        **kwargs (dict): All remaining properties are passed to the constructor of `LetTradeLive`

    Returns:
        LetTradeCCXT: _description_
    """
    api_kwargs: dict = kwargs.setdefault("api_kwargs", {})
    api_kwargs.update(
        exchange=ccxt_exchange,
        key=ccxt_key,
        secret=ccxt_secret,
        type=ccxt_type,
        verbose=ccxt_verbose,
    )

    return let_live(
        strategy=strategy,
        datas=datas,
        feeder=feeder,
        exchange=exchange,
        account=account,
        commander=commander,
        plotter=plotter,
        stats=stats,
        bot=bot,
        lettrade=lettrade,
        api=api,
        **kwargs,
    )



from typing import TYPE_CHECKING

import pandas as pd

# import plotly.graph_objects as go
from plotly import express as px

from ..plot import OptimizePlotter

if TYPE_CHECKING:
    from rich.progress import Progress


class PlotlyOptimizePlotter(OptimizePlotter):
    _process_bar: "Progress | None"

    def __init__(self, total=None, process_bar: bool = True) -> None:
        super().__init__()

        self._total = total

        if process_bar:
            from rich.progress import (
                BarColumn,
                Console,
                Progress,
                SpinnerColumn,
                TaskProgressColumn,
                TextColumn,
                TimeElapsedColumn,
                TimeRemainingColumn,
            )

            console = Console(record=True, force_jupyter=False)
            self._process_bar = Progress(
                SpinnerColumn(),
                # *Progress.get_default_columns(),
                TextColumn("[progress.description]{task.description}"),
                BarColumn(),
                TaskProgressColumn(),
                TextColumn("[bold blue][{task.completed}/{task.total}]"),
                TimeRemainingColumn(),
                TimeElapsedColumn(),
                console=console,
                # transient=False,
            )
            self._process_bar.add_task("[cyan2]Optimizing", total=total)
            self._process_bar.start()
        else:
            self._process_bar = None

    def on_result(self, result):
        if self._process_bar is not None:
            task_id = next(iter(self._process_bar._tasks))
            self._process_bar.update(
                task_id,
                advance=1,
                refresh=True,
            )

    def on_done(self):
        if self._process_bar is not None:
            # self._process_bar.refresh()
            self._process_bar.stop()

    def load(self):
        raise NotImplementedError

    def plot(self, **kwargs):
        """Plot optimize result"""
        ids = []
        equities = []
        for result in self.results:
            ids.append(result["index"])
            equities.append(result["result"]["equity"])

        df = pd.DataFrame({"id": ids, "equity": equities})

        fig = px.scatter(df, x="id", y="equity")
        fig.show()

    def stop(self):
        raise NotImplementedError

    def _xyzs(self, x: str, y: str, z: str):
        xs = []
        ys = []
        zs = []
        for result in self.results:
            xs.append(result["optimize"][x])
            ys.append(result["optimize"][y])
            zs.append(result["result"][z])

        return {x: xs, y: ys, z: zs}

    def _xyz_default(self, x, y, z):
        if x is None or y is None:
            if len(self.results) == 0:
                raise RuntimeError("Result is empty")

            optimize_keys = list(self.results[0]["optimize"].keys())
            if x is None:
                x = optimize_keys[0]
            if y is None:
                y = optimize_keys[1]
        if z is None:
            z = "equity"
        return x, y, z

    def heatmap(
        self,
        x: str = None,
        y: str = None,
        z: str = "equity",
        histfunc="max",
        **kwargs,
    ):
        """Plot optimize heatmap

        Args:
            x (str, optional): _description_. Defaults to None.
            y (str, optional): _description_. Defaults to None.
            z (str, optional): _description_. Defaults to "equity".
            histfunc (str, optional): _description_. Defaults to "max".
        """
        x, y, z = self._xyz_default(x, y, z)
        df = pd.DataFrame(self._xyzs(x=x, y=y, z=z))
        fig = px.density_heatmap(
            df,
            x=x,
            y=y,
            z=z,
            nbinsx=int(df[x].max() - df[x].min() + 1),
            nbinsy=int(df[y].max() - df[y].min() + 1),
            histfunc=histfunc,
            color_continuous_scale="Viridis",
            **kwargs,
        )
        fig.show()

    def contour(
        self,
        x: str = None,
        y: str = None,
        z: str = "equity",
        histfunc="max",
        **kwargs,
    ):
        """Plot optimize contour

        Args:
            x (str, optional): _description_. Defaults to None.
            y (str, optional): _description_. Defaults to None.
            z (str, optional): _description_. Defaults to "equity".
            histfunc (str, optional): _description_. Defaults to "max".
        """
        x, y, z = self._xyz_default(x, y, z)
        df = pd.DataFrame(self._xyzs(x=x, y=y, z=z))
        fig = px.density_contour(
            df,
            x=x,
            y=y,
            z=z,
            nbinsx=int(df[x].max() - df[x].min() + 1),
            nbinsy=int(df[y].max() - df[y].min() + 1),
            histfunc=histfunc,
            **kwargs,
        )
        fig.update_traces(contours_coloring="fill", contours_showlabels=True)
        fig.show()



from lettrade import BotStatistic, Commander, LetTrade, LetTradeBot, Plotter
from lettrade.strategy.strategy import Strategy

from .api import LiveAPI
from .data import LiveDataFeed


class LetTradeLiveBot(LetTradeBot):
    """Live bot object"""

    datas: list[LiveDataFeed]

    _api: LiveAPI = None

    def __init__(self, api: LiveAPI | None = LiveAPI, **kwargs) -> None:
        """_summary_

        Args:
            api (LiveAPI | None, optional): _description_. Defaults to LiveAPI.
        """
        super().__init__(**kwargs)

        if issubclass(api, LiveAPI):
            self._api = api(**self._kwargs.get("api_kwargs", {}))
        else:
            self._api = api

        self._kwargs.setdefault("feeder_kwargs", dict()).update(api=self._api)
        self._kwargs.setdefault("exchange_kwargs", dict()).update(api=self._api)
        self._kwargs.setdefault("account_kwargs", dict()).update(api=self._api)

        for data in self.datas:
            data._api = self._api

    def init(self):
        self._api.init()
        super().init()


class LetTradeLive(LetTrade):
    """Help to maintain live bots"""

    _data_cls: type[LiveDataFeed] = LiveDataFeed

    def _multiprocess(self, **kwargs):
        # Impletement api dependencies and save to api_kwargs
        api: type[LiveAPI] = self._kwargs.get("api")
        api_kwargs = self._kwargs.setdefault("api_kwargs", {})

        api_cls = api if issubclass(api, LiveAPI) else api.__class__
        api_cls.multiprocess(kwargs=api_kwargs)

        super()._multiprocess(**kwargs)

    def _datafeed(
        self,
        data: LiveDataFeed | list | set | tuple,
        **kwargs,
    ) -> LiveDataFeed:
        if isinstance(data, (list | set | tuple)):
            if len(data) < 2:
                raise RuntimeError("LiveDataFeed missing (symbol, timeframe)")
            symbol, timeframe = data[0], data[1]
            name = data[2] if len(data) > 2 else None

            data = self._data_cls(
                name=name,
                symbol=symbol,
                timeframe=timeframe,
            )
        elif isinstance(data, dict):
            data = self._data_cls(
                symbol=data.get("symbol"),
                timeframe=data.get("timeframe"),
                name=data.get("name", None),
            )
        elif isinstance(data, self._data_cls):
            pass
        else:
            return RuntimeError(f"Data {data} is not support yet")

        return super()._datafeed(data=data, **kwargs)


def let_live(
    datas: set[set[str]],
    strategy: type[Strategy],
    *,
    commander: Commander | None = None,
    stats: type[BotStatistic] = BotStatistic,
    plotter: type[Plotter] | None = None,
    bot: type[LetTradeLiveBot] = LetTradeLiveBot,
    lettrade: type[LetTradeLive] = LetTradeLive,
    api: type[LiveAPI] = LiveAPI,
    **kwargs,
) -> "LetTradeLive":
    """Help to build `LetTradeLive`

    Args:
        datas (set[set[str]]): _description_
        strategy (type[Strategy]): _description_
        commander (Commander | None, optional): _description_. Defaults to None.
        stats (type[BotStatistic], optional): _description_. Defaults to BotStatistic.
        plotter (type[Plotter] | None, optional): _description_. Defaults to None.
        bot (type[LetTradeLiveBot], optional): _description_. Defaults to LetTradeLiveBot.
        lettrade (type[LetTradeLive], optional): _description_. Defaults to LetTradeLive.
        api (type[LiveAPI], optional): _description_. Defaults to LiveAPI.

    Returns:
        LetTradeLive: _description_
    """
    return lettrade(
        strategy=strategy,
        datas=datas,
        commander=commander,
        plotter=plotter,
        stats=stats,
        bot=bot,
        api=api,
        **kwargs,
    )



from typing import Callable

import pandas as pd

DATAFRAME_PLOTTERS_NAME = "_lt_plotters"


class IndicatorPlotter:
    """Add indicator plotter to DataFrame"""

    def __init__(
        self,
        dataframe: pd.DataFrame,
        plotter: Callable,
        filter: Callable | pd.Series | None = None,
        push: bool = True,
        **kwargs,
    ) -> None:
        # self.dataframe: pd.DataFrame = dataframe
        self.plotter: Callable = plotter
        self.filter: Callable | pd.Series | None = filter
        self.kwargs = kwargs

        if push:
            indicator_push_plotter(dataframe=dataframe, ip=self)

    def config(self, dataframe: pd.DataFrame):
        df_name = dataframe.name
        if self.filter is not None:
            if callable(self.filter):
                filter = self.filter(dataframe)
            else:
                filter = self.filter

            self.kwargs.update(filter=filter)

        config = self.plotter(dataframe=dataframe, **self.kwargs)

        if df_name in config:
            return config

        return {f"{df_name}": config}


def indicator_push_plotter(dataframe: pd.DataFrame, ip: IndicatorPlotter):
    if not hasattr(dataframe, DATAFRAME_PLOTTERS_NAME):
        object.__setattr__(dataframe, DATAFRAME_PLOTTERS_NAME, [])

    getattr(dataframe, DATAFRAME_PLOTTERS_NAME).append(ip)


def indicator_load_plotters(dataframe: pd.DataFrame) -> dict:
    if not hasattr(dataframe, DATAFRAME_PLOTTERS_NAME):
        return dict()

    from lettrade.plot.plotly import plot_merge

    config = dict()
    for plotter in getattr(dataframe, DATAFRAME_PLOTTERS_NAME):
        plot_merge(config, plotter.config(dataframe=dataframe))

    return config



import pandas as pd


def series_init(
    series: pd.Series | str = "close",
    dataframe: pd.DataFrame | None = None,
    inplace: bool = False,
):

    if dataframe is None:
        if __debug__:
            if not isinstance(series, pd.Series):
                raise RuntimeError(
                    f"Series type '{type(series)}' is not instance of pandas.Series"
                )

            if inplace:
                raise RuntimeError("DataFrame isnot set when inplace=True")
    else:
        if __debug__:
            if not isinstance(dataframe, pd.DataFrame):
                raise RuntimeError(
                    f"DataFrame type '{type(dataframe)}' "
                    "is not instance of pandas.DataFrame"
                )

            if not isinstance(series, str):
                raise RuntimeError(
                    f"Series type {type(series)} is not string of column name"
                )

        series = dataframe[series]
    return series


def series_indicator_inject(fn):
    def __call(data, *args, **kwargs):
        if isinstance(data, pd.DataFrame):
            return fn(*args, **kwargs, dataframe=data)

        if isinstance(data, pd.Series):
            return fn(data, *args, **kwargs)

        raise RuntimeError(f"Indicator parameter type {type(data)} is invalid")

    return __call


def pandas_inject(obj: object | None = None):
    if obj is None:
        from pandas.core.base import PandasObject

        obj = PandasObject

    obj.diff = series_indicator_inject(diff)
    obj.above = series_indicator_inject(above)
    obj.below = series_indicator_inject(below)
    obj.direction = series_indicator_inject(direction)
    obj.rolling_above = series_indicator_inject(rolling_above)
    obj.rolling_below = series_indicator_inject(rolling_below)
    obj.rolling_direction = series_indicator_inject(rolling_direction)
    obj.rolling_min = series_indicator_inject(rolling_min)
    obj.rolling_max = series_indicator_inject(rolling_max)
    obj.rolling_mean = series_indicator_inject(rolling_mean)
    obj.crossover = series_indicator_inject(crossover)
    obj.crossunder = series_indicator_inject(crossunder)


def diff(
    series1: pd.Series | str,
    series2: pd.Series | str,
    dataframe: pd.DataFrame = None,
    name: str | None = None,
    prefix: str = "",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_kwargs: dict | None = None,
    # **kwargs,
) -> pd.Series:
    """Difference between 2 series

    Args:
        series1 (pd.Series): _description_
        series2 (pd.Series): _description_
        dataframe (pd.DataFrame, optional): _description_. Defaults to None.
        name (str | None, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "".
        inplace (bool, optional): _description_. Defaults to False.
        plot (bool | list, optional): _description_. Defaults to False.
        plot_kwargs (dict | None, optional): _description_. Defaults to None.

    Raises:
        RuntimeError: _description_

    Returns:
        pd.Series: Difference of 2 series (series1 - series2)
    """
    if __debug__:
        if plot and not inplace:
            raise RuntimeError("Cannot plot when inplace=False")

    if isinstance(series1, str):
        series1 = dataframe[series1]
    if isinstance(series2, str):
        series2 = dataframe[series2]

    i = series1 - series2

    if inplace:
        name = name or f"{prefix}diff"
        dataframe[name] = i

        if plot:
            _plot_mark(dataframe=dataframe, name=name, plot_kwargs=plot_kwargs)

    return i


def s_above(series1: pd.Series, series2: pd.Series) -> pd.Series:
    return (series1 - series2).apply(lambda v: 100 if v > 0 else 0)


def s_below(series1: pd.Series, series2: pd.Series) -> pd.Series:
    return (series1 - series2).apply(lambda v: -100 if v < 0 else 0)


def s_direction(series1: pd.Series, series2: pd.Series) -> pd.Series:
    return (series1 - series2).apply(lambda v: -100 if v < 0 else 100 if v > 0 else 0)


def above(
    series1: pd.Series | str,
    series2: pd.Series | str,
    dataframe: pd.DataFrame = None,
    name: str | None = None,
    prefix: str = "",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_kwargs: dict | None = None,
    # **kwargs,
) -> pd.Series:
    """Check a Series is above another Series

    Args:
        series1 (pd.Series): first Series
        series2 (pd.Series): second Series

    Returns:
        pd.Series: True series1 is above series2 else False
    """
    if __debug__:
        if plot and not inplace:
            raise RuntimeError("Cannot plot when inplace=False")

    if isinstance(series1, str):
        series1 = dataframe[series1]
    if isinstance(series2, str):
        series2 = dataframe[series2]

    i = s_above(series1=series1, series2=series2)

    if inplace:
        name = name or f"{prefix}above"
        dataframe[name] = i

        if plot:
            _plot_mark(dataframe=dataframe, name=name, plot_kwargs=plot_kwargs)

    return i


def below(
    series1: pd.Series | str,
    series2: pd.Series | str,
    dataframe: pd.DataFrame = None,
    name: str | None = None,
    prefix: str = "",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_kwargs: dict | None = None,
    # **kwargs,
) -> pd.Series:
    """Check a Series is below another Series

    Args:
        series1 (pd.Series): first Series
        series2 (pd.Series): second Series

    Returns:
        pd.Series: -100 series1 is below series2 else 0
    """
    if __debug__:
        if plot and not inplace:
            raise RuntimeError("Cannot plot when inplace=False")

    if isinstance(series1, str):
        series1 = dataframe[series1]
    if isinstance(series2, str):
        series2 = dataframe[series2]

    i = s_below(series1, series2)

    if inplace:
        name = name or f"{prefix}below"
        dataframe[name] = i

        if plot:
            _plot_mark(dataframe=dataframe, name=name, plot_kwargs=plot_kwargs)

    return i


def direction(
    series1: pd.Series | str,
    series2: pd.Series | str,
    dataframe: pd.DataFrame = None,
    name: str | None = None,
    prefix: str = "",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_kwargs: dict | None = None,
    # **kwargs,
) -> pd.Series:
    """Check a Series is direction another Series

    Args:
        series1 (pd.Series): first Series
        series2 (pd.Series): second Series

    Returns:
        pd.Series: -100 if series1 < series2, 100 if series1 > series2, else 0
    """
    if __debug__:
        if plot and not inplace:
            raise RuntimeError("Cannot plot when inplace=False")

    if isinstance(series1, str):
        series1 = dataframe[series1]
    if isinstance(series2, str):
        series2 = dataframe[series2]

    i = s_direction(series1, series2)

    if inplace:
        name = name or f"{prefix}direction"
        dataframe[name] = i

        if plot:
            _plot_mark(dataframe=dataframe, name=name, plot_kwargs=plot_kwargs)

    return i


def rolling_above(
    series1: pd.Series,
    series2: pd.Series,
    window: int = 20,
    min_periods: int | None = None,
    dataframe: pd.DataFrame = None,
    name: str | None = None,
    prefix: str = "",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_kwargs: dict | None = None,
    # **kwargs,
) -> pd.Series:
    """Check a Series is rolling above another Series

    Args:
        series1 (pd.Series): first Series
        series2 (pd.Series): second Series

    Returns:
        pd.Series: 100 mean series1 is rolling above series2 else 0
    """
    if __debug__:
        if plot and not inplace:
            raise RuntimeError("Cannot plot when inplace=False")

    if isinstance(series1, str):
        series1 = dataframe[series1]
    if isinstance(series2, str):
        series2 = dataframe[series2]

    min_periods = window if min_periods is None else min_periods

    i = s_above(series1, series2)
    i = i.rolling(window=window, min_periods=min_periods).min()
    i = i.apply(lambda v: 100 if v >= 100 else 0).astype(int)

    if inplace:
        name = name or f"{prefix}rolling_above"
        dataframe[name] = i

        if plot:
            _plot_mark(dataframe=dataframe, name=name, plot_kwargs=plot_kwargs)

    return i


def rolling_below(
    series1: pd.Series,
    series2: pd.Series,
    window: int = 20,
    min_periods: int | None = None,
    dataframe: pd.DataFrame = None,
    name: str | None = None,
    prefix: str = "",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_kwargs: dict | None = None,
    # **kwargs,
) -> pd.Series:
    """Check a Series is rolling below another Series

    Args:
        series1 (pd.Series): first Series
        series2 (pd.Series): second Series

    Returns:
        pd.Series: -100 mean series1 is rolling below series2 else 0
    """
    if __debug__:
        if plot and not inplace:
            raise RuntimeError("Cannot plot when inplace=False")

    if isinstance(series1, str):
        series1 = dataframe[series1]
    if isinstance(series2, str):
        series2 = dataframe[series2]

    min_periods = window if min_periods is None else min_periods

    i = s_below(series1, series2)
    i = i.rolling(window=window, min_periods=min_periods).max()
    i = i.apply(lambda v: -100 if v <= -100 else 0).astype(int)

    if inplace:
        name = name or f"{prefix}rolling_below"
        dataframe[name] = i

        if plot:
            _plot_mark(dataframe=dataframe, name=name, plot_kwargs=plot_kwargs)

    return i


def rolling_direction(
    series1: pd.Series,
    series2: pd.Series,
    window: int = 20,
    min_periods: int | None = None,
    dataframe: pd.DataFrame = None,
    name: str | None = None,
    prefix: str = "",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_kwargs: dict | None = None,
    # **kwargs,
) -> pd.Series:
    """Check a Series is rolling on one side with another Series

    Args:
        series1 (pd.Series): first Series
        series2 (pd.Series): second Series

    Returns:
        pd.Series:  100 mean series1 keep consistences above series2 at least `window` bars
                    -100 mean series1 keep consistences below series2 at least `window` bars
                    0 else cases
    """
    if __debug__:
        if plot and not inplace:
            raise RuntimeError("Cannot plot when inplace=False")

    if isinstance(series1, str):
        series1 = dataframe[series1]
    if isinstance(series2, str):
        series2 = dataframe[series2]

    min_periods = window if min_periods is None else min_periods

    i = s_direction(series1, series2)
    i = i.rolling(window=window, min_periods=min_periods).mean()
    i = i.apply(lambda v: -100 if v <= -100 else 100 if v >= 100 else 0).astype(int)

    if inplace:
        name = name or f"{prefix}rolling_direction"
        dataframe[name] = i

        if plot:
            _plot_mark(dataframe=dataframe, name=name, plot_kwargs=plot_kwargs)

    return i


def rolling_min(
    series: pd.Series,
    window: int = 14,
    min_periods: int | None = None,
    dataframe: pd.DataFrame = None,
    name: str | None = None,
    prefix: str = "",
    inplace: bool = False,
):
    if isinstance(series, str):
        series = dataframe[series]

    min_periods = window if min_periods is None else min_periods
    i = series.rolling(window=window, min_periods=min_periods).min()

    if inplace:
        name = name or f"{prefix}rolling_min"
        dataframe[name] = i

    return i


def rolling_max(
    series: pd.Series,
    window: int = 14,
    min_periods: int | None = None,
    dataframe: pd.DataFrame = None,
    name: str | None = None,
    prefix: str = "",
    inplace: bool = False,
):
    if isinstance(series, str):
        series = dataframe[series]

    min_periods = window if min_periods is None else min_periods
    i = series.rolling(window=window, min_periods=min_periods).max()

    if inplace:
        name = name or f"{prefix}rolling_max"
        dataframe[name] = i

    return i


def rolling_mean(
    series: pd.Series,
    window: int = 14,
    min_periods: int | None = None,
    dataframe: pd.DataFrame = None,
    name: str | None = None,
    prefix: str = "",
    inplace: bool = False,
):
    if isinstance(series, str):
        series = dataframe[series]

    min_periods = window if min_periods is None else min_periods
    i = series.rolling(window=window, min_periods=min_periods).mean()

    if inplace:
        name = name or f"{prefix}rolling_mean"
        dataframe[name] = i

    return i


def crossover(
    series1: pd.Series | str,
    series2: pd.Series | str,
    dataframe: pd.DataFrame = None,
    name: str | None = None,
    prefix: str = "",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_kwargs: dict | None = None,
    # **kwargs,
) -> pd.Series:
    """Check if a Series cross over another Series

    Args:
        series1 (pd.Series): first Series
        series2 (pd.Series): second Series

    Returns:
        pd.Series: 100 if series1 cross over series2 else 0
    """
    if __debug__:
        if plot and not inplace:
            raise RuntimeError("Cannot plot when inplace=False")

    if isinstance(series1, str):
        series1 = dataframe[series1]
    if isinstance(series2, str):
        series2 = dataframe[series2]

    below1 = s_below(series1, series2).shift(1)
    above0 = s_above(series1, series2)
    i = (-below1 + above0).apply(lambda v: 100 if v >= 200 else 0)

    if inplace:
        name = name or f"{prefix}crossover"
        dataframe[name] = i

        if plot:
            _plot_mark(dataframe=dataframe, name=name, plot_kwargs=plot_kwargs)

    return i


def crossunder(
    series1: pd.Series | str,
    series2: pd.Series | str,
    dataframe: pd.DataFrame = None,
    name: str | None = None,
    prefix: str = "",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_kwargs: dict | None = None,
    # **kwargs,
) -> pd.Series:
    """Check if a Series cross under another Series

    Args:
        series1 (pd.Series): first Series
        series2 (pd.Series): second Series

    Returns:
        pd.Series: -100 if series1 cross under series2 else 0
    """
    if __debug__:
        if plot and not inplace:
            raise RuntimeError("Cannot plot when inplace=False")

    if isinstance(series1, str):
        series1 = dataframe[series1]
    if isinstance(series2, str):
        series2 = dataframe[series2]

    above1 = s_above(series1, series2).shift(1)
    below0 = s_below(series1, series2)
    i = (below0 - above1).apply(lambda v: -100 if v <= -200 else 0)

    if inplace:
        name = name or f"{prefix}crossunder"
        dataframe[name] = i

        if plot:
            _plot_mark(dataframe=dataframe, name=name, plot_kwargs=plot_kwargs)

    return i


def _plot_mark(
    dataframe: pd.DataFrame = None,
    name: str | None = None,
    plot_kwargs: dict | None = None,
):
    if plot_kwargs is None:
        plot_kwargs = dict()

    plot_kwargs.update(series="close", name=name)

    from lettrade.indicator.plot import IndicatorPlotter
    from lettrade.plot.plotly import plot_mark

    IndicatorPlotter(
        dataframe=dataframe,
        plotter=plot_mark,
        filter=lambda df: df[name] != 0,
        **plot_kwargs,
    )



from sys import float_info as sflt

import numpy as np
import pandas as pd


def zero(x: tuple[int, float]) -> tuple[int, float]:
    """If the value is close to zero, then return zero. Otherwise return itself."""
    return 0 if abs(x) < sflt.epsilon else x


def parabolic_sar(
    dataframe: pd.DataFrame,
    af0: float = 0.02,
    af: float = 0.02,
    max_af: float = 0.2,
    prefix: str = "psar_",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_kwargs: dict | None = None,
) -> dict[str, pd.Series] | pd.DataFrame:
    """Indicator: Parabolic Stop and Reverse (PSAR)

    Source:
        https://github.com/twopirllc/pandas-ta/blob/main/pandas_ta/trend/psar.py

    Args:
        dataframe (pd.DataFrame): _description_
        af0 (float, optional): _description_. Defaults to 0.02.
        af (float, optional): _description_. Defaults to 0.02.
        max_af (float, optional): _description_. Defaults to 0.2.
        prefix (str, optional): _description_. Defaults to "psar_".
        inplace (bool, optional): _description_. Defaults to False.
        plot (bool | list, optional): _description_. Defaults to False.
        plot_kwargs (dict | None, optional): _description_. Defaults to None.

    Raises:
        RuntimeError: _description_

    Returns:
        dict[str, pd.Series] | pd.DataFrame: _description_
    """
    if __debug__:
        if not isinstance(dataframe, pd.DataFrame):
            raise RuntimeError(
                f"dataframe type '{type(dataframe)}' "
                "is not instance of pandas.DataFrame"
            )

    def _falling(high: pd.Series, low: pd.Series, drift: int = 1):
        """Returns the last -DM value"""
        # Not to be confused with ta.falling()
        up = high - high.shift(drift)
        dn = low.shift(drift) - low
        dmn = (((dn > up) & (dn > 0)) * dn).apply(zero).iloc[-1]
        return dmn > 0

    # Falling if the first NaN -DM is positive
    falling = _falling(dataframe.high.iloc[:2], dataframe.low.iloc[:2])
    if falling:
        sar = dataframe.high.iloc[0]
        ep = dataframe.low.iloc[0]
    else:
        sar = dataframe.low.iloc[0]
        ep = dataframe.high.iloc[0]

    sar = dataframe.close.iloc[0]

    i_long = pd.Series(np.nan, index=dataframe.high.index)
    i_short = i_long.copy()
    i_reversal = pd.Series(0, index=dataframe.high.index)
    i_af = i_long.copy()
    i_af.iloc[0:2] = af0

    # Calculate Result
    m = dataframe.high.shape[0]
    for row in range(1, m):
        high_ = dataframe.high.iloc[row]
        low_ = dataframe.low.iloc[row]

        if falling:
            _sar = sar + af * (ep - sar)
            reverse = high_ > _sar

            if low_ < ep:
                ep = low_
                af = min(af + af0, max_af)

            _sar = max(dataframe.high.iloc[row - 1], dataframe.high.iloc[row - 2], _sar)
        else:
            _sar = sar + af * (ep - sar)
            reverse = low_ < _sar

            if high_ > ep:
                ep = high_
                af = min(af + af0, max_af)

            _sar = min(dataframe.low.iloc[row - 1], dataframe.low.iloc[row - 2], _sar)

        if reverse:
            _sar = ep
            af = af0
            falling = not falling  # Must come before next line
            ep = low_ if falling else high_

        sar = _sar  # Update SAR

        # Seperate long/short sar based on falling
        if falling:
            i_short.iloc[row] = sar
        else:
            i_long.iloc[row] = sar

        i_af.iloc[row] = af
        i_reversal.iloc[row] = int(reverse)

    # Result is inplace or new dict
    result = dataframe if inplace else {}

    result[f"{prefix}long"] = i_long
    result[f"{prefix}short"] = i_short
    result[f"{prefix}af"] = i_af
    result[f"{prefix}reversal"] = i_reversal

    # Plot
    if plot:
        if plot_kwargs is None:
            plot_kwargs = dict()

        if isinstance(plot, list):
            if f"{prefix}long" in plot:
                plot_kwargs.update(long=f"{prefix}long")
            if f"{prefix}short" in plot:
                plot_kwargs.update(short=f"{prefix}short")
        else:
            plot_kwargs.update(
                long=f"{prefix}long",
                short=f"{prefix}short",
            )

        from lettrade.indicator.plot import IndicatorPlotter
        from lettrade.plot.plotly import plot_parabolic_sar

        IndicatorPlotter(
            dataframe=dataframe,
            plotter=plot_parabolic_sar,
            **plot_kwargs,
        )

    return result



import pandas as pd
import talib.abstract as ta

from ..series import series_init


def rsi(
    series: pd.Series | str = "close",
    window: int = None,
    dataframe: pd.DataFrame | None = None,
    name: str | None = None,
    prefix: str = "",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_kwargs: dict | None = None,
    **kwargs,
) -> pd.Series | pd.DataFrame:
    """_summary_

    Args:
        series (pd.Series | str, optional): _description_. Defaults to "close".
        window (int, optional): _description_. Defaults to None.
        dataframe (pd.DataFrame, optional): _description_. Defaults to None.
        name (str | None, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "".
        inplace (bool, optional): _description_. Defaults to False.
        plot (bool | list, optional): _description_. Defaults to False.
        plot_kwargs (dict | None, optional): _description_. Defaults to None.

    Raises:
        RuntimeError: _description_

    Returns:
        pd.Series | pd.DataFrame: {rsi}
    """
    # Validation & init
    if __debug__:
        if window is None or window <= 0:
            raise RuntimeError(f"Window {window} is invalid")
    series = series_init(series=series, dataframe=dataframe, inplace=inplace)

    # Indicator
    i = ta.RSI(series, timeperiod=window, **kwargs)

    if inplace:
        name = name or f"{prefix}rsi"
        dataframe[name] = i

        # Plot
        if plot:
            if plot_kwargs is None:
                plot_kwargs = dict()

            plot_kwargs.update(series=name, name=name)
            plot_kwargs.setdefault("row", 2)
            plot_kwargs.setdefault("row_height", 0.5)

            from lettrade.indicator.plot import IndicatorPlotter
            from lettrade.plot.plotly import plot_line

            IndicatorPlotter(dataframe=dataframe, plotter=plot_line, **plot_kwargs)

        return dataframe

    return i



"""Import everything in one. Import some unnecessary but convenient for everyone"""

from . import *
from .commander.telegram import *
from .exchange.backtest import *
from .exchange.backtest.plotly import *
from .exchange.ccxt import *
from .exchange.metatrader import *
from .indicator import *
from .plot.plotly import *



import logging
from pathlib import Path

import pandas as pd
import yfinance as yf

from lettrade.data.extra.csv import csv_export

logger = logging.getLogger(__name__)


def yf_download(tickers, path=None, force=False, interval="1d", *args, **kwargs):
    # path validate
    if path is None:
        path = f"data/{tickers}_{interval}.csv"

    path = Path(path)
    if not force and path.exists():
        logger.error("File %s existed", path)
        return

    # Download
    df: pd.DataFrame = yf.download(tickers=tickers, interval=interval, *args, **kwargs)
    logger.info("YFinance downloaded:\n%s\n%s", df.head(), df.tail())

    # Parse
    df = yf_parse(df)

    # Save to csv
    return csv_export(df, path=path)


def yf_parse(df: pd.DataFrame):
    df = df[df.High != df.Low]
    df.index = df.index.rename("datetime")
    df = df.rename(
        columns={
            "Open": "open",
            "High": "high",
            "Low": "low",
            "Close": "close",
            "Volume": "volume",
        }
    )
    df = df.filter(["datetime", "open", "high", "low", "close", "volume"], axis=1)
    return df



import logging
import math
import re
from datetime import datetime, timedelta

import pandas as pd

logger = logging.getLogger(__name__)


TIMEFRAME_UNIT_LET_2_PANDAS = {
    "s": "s",
    "m": "m",
    "h": "h",
    "d": "day",
    "w": "W",
    # "1mn"
}
TIMEFRAME_UNIT_PANDAS_2_LET = {v: k for k, v in TIMEFRAME_UNIT_LET_2_PANDAS.items()}

TIMEFRAME_DELTA_2_STR = {
    pd.Timedelta(1, "m"): (1, "m"),
    pd.Timedelta(2, "m"): (2, "m"),
    pd.Timedelta(3, "m"): (3, "m"),
    pd.Timedelta(4, "m"): (4, "m"),
    pd.Timedelta(5, "m"): (5, "m"),
    pd.Timedelta(6, "m"): (6, "m"),
    pd.Timedelta(10, "m"): (10, "m"),
    pd.Timedelta(12, "m"): (12, "m"),
    pd.Timedelta(15, "m"): (15, "m"),
    pd.Timedelta(20, "m"): (20, "m"),
    pd.Timedelta(30, "m"): (30, "m"),
    pd.Timedelta(1, "h"): (1, "h"),
    pd.Timedelta(2, "h"): (2, "h"),
    pd.Timedelta(3, "h"): (3, "h"),
    pd.Timedelta(4, "h"): (4, "h"),
    pd.Timedelta(6, "h"): (6, "h"),
    pd.Timedelta(8, "h"): (8, "h"),
    pd.Timedelta(12, "h"): (12, "h"),
    pd.Timedelta(1, "day"): (1, "d"),
    pd.Timedelta(1, "W"): (1, "w"),
    # "1mn": pd.Timedelta(days=30),
}

_pattern_timeframe_str = re.compile(r"^([0-9]+)([a-z]+)$")


class TimeFrame:
    """DataFeed TimeFrame"""

    delta: pd.Timedelta
    unit: str
    unit_pandas: str
    value: int

    def __init__(self, tf: "int | str | list | pd.Timedelta | TimeFrame") -> None:
        """_summary_

        Args:
            tf (int | str | list | pd.Timedelta | TimeFrame):
                - `int`: TimeFrame in minutes. Example: TimeFrame(5) == TimeFrame("5m")
                - `str`: string format of TimeFrame in `s`, `m`, `h`, `d`, `w`.
        Raises:
            RuntimeError: _description_
        """
        if isinstance(tf, TimeFrame):
            self.value = tf.value
            self.unit = tf.unit
        elif isinstance(tf, str):
            match = _pattern_timeframe_str.search(tf)
            if not match:
                raise RuntimeError(f"TimeFrame value {tf} is invalid")

            self.value = int(match.group(1))
            self.unit = match.group(2)
        elif isinstance(tf, int):
            self.value = tf
            self.unit = "m"
        elif isinstance(tf, list):
            self.value = int(tf[0])
            self.unit = tf[1]
        elif isinstance(tf, pd.Timedelta):
            map = TIMEFRAME_DELTA_2_STR[tf]
            self.value = map[0]
            self.unit = map[1]
        else:
            raise RuntimeError(f"Timeframe {tf} is invalid format")

        # Validate
        self._validate()

        # Setup
        self.unit_pandas = TIMEFRAME_UNIT_LET_2_PANDAS[self.unit]
        self.delta = pd.Timedelta(self.value, self.unit_pandas)

        # Warning
        if self.delta not in TIMEFRAME_DELTA_2_STR:
            if self.unit not in ["s", "m"]:
                logger.warning(
                    "Unsupport TimeFrame(%s), some function may not work floor()/ceil()...",
                    self.delta,
                )

    def _validate(self):
        # Unit
        if self.unit not in TIMEFRAME_UNIT_LET_2_PANDAS:
            raise RuntimeError(f"TimeFrame unit {self.unit} is invalid")

        # Value
        if self.value <= 0:
            raise RuntimeError(f"Timeframe value {self.value} is <= 0")
        elif self.unit in ["s", "m"] and self.value >= 60:
            raise RuntimeError(f"Timeframe value {self.value} is >= 60{self.unit}")
        elif self.unit == "h" and self.value >= 24:
            raise RuntimeError(f"Timeframe value {self.value} is >= 24h")
        elif self.unit == "d" and self.value >= 7:
            raise RuntimeError(f"Timeframe value {self.value} is >= 7d")

    def __repr__(self) -> str:
        return self.string

    @property
    def string(self):
        """TimeFrame as lettrade string"""
        return f"{self.value}{self.unit}"

    @property
    def string_pandas(self):
        """TimeFrame as pandas string"""
        return f"{self.value}{self.unit_pandas}"

    def floor(
        self, at: datetime | timedelta | pd.Timestamp | pd.Timedelta
    ) -> pd.Timestamp | pd.Timedelta:
        """Get floor of TimeFrame

        Args:
            at (datetime | timedelta | pd.Timestamp | pd.Timedelta): _description_

        Raises:
            RuntimeError: _description_

        Returns:
            pd.Timestamp | pd.Timedelta: _description_
        """
        if isinstance(at, datetime):
            at = pd.Timestamp(at)
        elif isinstance(at, timedelta):
            at = pd.Timedelta(at)

        if self.unit in ["h", "m", "s"]:
            freq = self.string_pandas
            if self.unit == "m":
                freq += "in"
            return at.floor(freq=freq)

        if isinstance(at, pd.Timestamp):
            if self.unit == "d":
                return pd.Timestamp(at.date())
            if self.unit == "w":
                return pd.Timestamp(at.date() - pd.Timedelta(days=at.day_of_week))

        elif isinstance(at, pd.Timedelta):
            if self.unit == "d":
                return pd.Timedelta(days=at.days)
            if self.unit == "w":
                weeks = math.floor(at.days / 7)
                return pd.Timedelta(weeks=weeks)

        raise RuntimeError(f"Unit {self.unit} is not implement yet for {at}")

    def ceil(
        self, at: datetime | timedelta | pd.Timestamp | pd.Timedelta
    ) -> pd.Timestamp | pd.Timedelta:
        """Get ceil of TimeFrame

        Args:
            at (datetime | timedelta | pd.Timestamp | pd.Timedelta): _description_

        Raises:
            RuntimeError: _description_

        Returns:
            pd.Timestamp: _description_
        """
        if isinstance(at, datetime):
            at = pd.Timestamp(at)
        elif isinstance(at, timedelta):
            at = pd.Timedelta(at)

        if self.unit in ["h", "m", "s"]:
            freq = self.string_pandas
            if self.unit == "m":
                freq += "in"
            return at.ceil(freq=freq)

        if isinstance(at, pd.Timestamp):
            if self.unit == "d":
                return pd.Timestamp(at.date() + pd.Timedelta(days=1))
            if self.unit == "w":
                return pd.Timestamp(at.date() + pd.Timedelta(days=7 - at.day_of_week))

        elif isinstance(at, pd.Timedelta):
            if self.unit == "d":
                return pd.Timedelta(days=at.days + (0 if at.seconds == 0 else 1))
            if self.unit == "w":
                weeks = math.ceil(at.days / 7)
                return pd.Timedelta(weeks=weeks)

        raise RuntimeError(f"Unit {self.unit} is not implement yet for {at}")



import logging
from typing import TYPE_CHECKING

import pandas as pd

from lettrade.exchange import (
    Execution,
    Order,
    OrderResult,
    OrderState,
    OrderType,
    Position,
    PositionResult,
    PositionState,
)

if TYPE_CHECKING:
    from .exchange import BackTestExchange

logger = logging.getLogger(__name__)


class BackTestExecution(Execution):
    """
    Execution for backtesting
    """

    def __init__(self, exchange: "BackTestExchange", *args, **kwargs):
        if exchange.executions is None:
            logger.warning(
                "Execution transaction is disable, enable by flag: show_execution=True"
            )
            return

        super().__init__(*args, exchange=exchange, **kwargs)

    @classmethod
    def from_order(
        cls,
        order: "BackTestOrder",
        price: float,
        at: object,
        size: float | None = None,
    ) -> "BackTestExecution":
        """Method help to build Execution object from Order object

        Args:
            price (float): Executed price
            at (object): Executed bar
            size (float | None, optional): Executed size. Defaults to None.

        Returns:
            BackTestExecution: Execution object
        """
        return cls(
            id=order.id,
            size=size or order.size,
            exchange=order.exchange,
            data=order.data,
            price=price,
            at=at,
            order=order,
        )


class BackTestOrder(Order):
    """Order for backtesting"""

    parent: "BackTestPosition"

    def cancel(
        self,
        caller: Order | Position | None = None,
        **kwargs,
    ) -> "OrderResult":
        """Cancel the Order and notify Exchange"""
        if self.state != OrderState.Placed:
            raise RuntimeError(f"Order {self.id} state {self.state} is not Placed")

        if self.parent:
            if self is self.parent.sl_order:
                self.parent.sl_order = None
            elif self is self.parent.tp_order:
                self.parent.tp_order = None

        return super().cancel()

    def fill(self, price: float, at: object, **kwargs) -> BackTestExecution:
        """Execution order and notify for Exchange

        Args:
            price (float): Executed price
            at (object): Executed bar

        Raises:
            RuntimeError: _description_

        Returns:
            BackTestExecution: Execution object
        """
        if self.state != OrderState.Placed:
            raise RuntimeError(f"Execution a {self.state} order")

        # Order
        ok = super().fill(price=price, at=at)

        # Execution is enable
        if self.exchange.executions is not None:
            execution = BackTestExecution.from_order(order=self, price=price, at=at)
            execution._on_execution()

        # Position hit SL/TP
        if self.parent:
            self.parent.exit(price=price, at=at, caller=self)
        else:
            # Position: Place and create new position
            position = BackTestPosition.from_order(order=self)

            position.entry(price=price, at=at)

        return ok

    @classmethod
    def from_position(
        cls,
        position: "BackTestPosition",
        id: str,
        type: OrderType,
        limit_price: float = None,
        stop_price: float = None,
        **kwargs,
    ) -> "BackTestOrder":
        order = cls(
            id=id,
            exchange=position.exchange,
            data=position.data,
            size=-position.size,
            type=type,
            limit_price=limit_price,
            stop_price=stop_price,
            tag=position.tag,
            parent=position,
            **kwargs,
        )
        return order


class BackTestPosition(Position):
    """Position for backtesting"""

    def update(
        self,
        sl: float | None = None,
        tp: float | None = None,
        **kwargs,
    ) -> PositionResult:
        if not sl and not tp:
            raise RuntimeError("Update sl=None and tp=None")

        if sl is not None:
            if self.sl_order:
                self.sl_order.update(stop_price=sl)
            else:
                self._new_sl_order(stop_price=sl)

        if tp is not None:
            if self.tp_order:
                self.tp_order.update(limit_price=tp)
            else:
                self._new_tp_order(limit_price=tp)

        return super().update(**kwargs)

    def entry(self, price: float, at: object, **kwargs) -> PositionResult:
        # Fee
        fee = self._account.fee(size=self.size)
        return super().entry(price=price, at=at, fee=fee, **kwargs)

    def exit(
        self,
        price: float | None = None,
        at: pd.Timestamp | None = None,
        caller: Order | Position | None = None,
        **kwargs,
    ) -> PositionResult:
        """Exit Position

        Args:
            price (float): Exit price
            at (object): Exit bar
            caller (Order | Position, optional): Skip caller to prevent infinite recursion loop. Defaults to None.
        """
        if self.state == PositionState.Exit:
            if caller is None:
                # Call by user
                raise RuntimeError(f"Call exited position {self}")
            return

        if caller is None:
            # Call by user
            if price is not None:
                raise RuntimeError(f"Price set {price} is not available")
            if at is not None:
                raise RuntimeError(f"At set {at} is not available")

            price = self.data.l.open[0]
            at = self.data.l.index[0]
        else:
            # Call by SL/TP order
            if price is None or at is None:
                raise RuntimeError(f"Caller {caller} with price is None or at is None")

        # PnL
        pl = self._account.pl(
            size=self.size,
            entry_price=self.entry_price,
            exit_price=price,
        )

        # Fee
        fee = self._account.fee(size=self.size)

        # State
        ok = super().exit(price=price, at=at, pl=pl, fee=fee, **kwargs)

        # Caller is position close by tp/sl order
        if caller is None:
            if self.sl_order is not None:
                self.sl_order.cancel(caller=self)
            if self.tp_order is not None:
                self.tp_order.cancel(caller=self)
        else:
            if self.sl_order and self.sl_order is not caller:
                self.sl_order.cancel(caller=caller)
            if self.tp_order and self.tp_order is not caller:
                self.tp_order.cancel(caller=caller)

        return ok

    def _new_sl_order(self, stop_price: float) -> BackTestOrder:
        if self.sl_order:
            raise RuntimeError(f"Position {self.id} SL Order {self.sl_order} existed")

        sl_order = BackTestOrder.from_position(
            position=self,
            id=f"{self.id}-sl",
            type=OrderType.Stop,
            stop_price=stop_price,
        )
        self.sl_order = sl_order
        sl_order.place(at=self.data.bar())
        return sl_order

    def _new_tp_order(self, limit_price: float) -> BackTestOrder:
        if self.tp_order:
            raise RuntimeError(f"Position {self.id} TP Order {self.tp_order} existed")

        tp_order = BackTestOrder.from_position(
            position=self,
            id=f"{self.id}-tp",
            type=OrderType.Limit,
            limit_price=limit_price,
        )

        self.tp_order = tp_order
        tp_order.place(at=self.data.bar())
        return tp_order

    @classmethod
    def from_order(
        cls,
        order: "BackTestOrder",
        size: float | None = None,
        state: PositionState = PositionState.Open,
        **kwargs,
    ) -> "BackTestPosition":
        """Build Position object from Order object

        Args:
            size (float, optional): Size of Position object. Defaults to None.
            state (PositionState, optional): State of Position object. Defaults to PositionState.Open.

        Returns:
            BackTestPosition: Position object
        """
        position = cls(
            id=order.id,
            size=size or order.size,
            exchange=order.exchange,
            data=order.data,
            state=state,
            parent=order,
        )
        if order.sl_price:
            position._new_sl_order(stop_price=order.sl_price)
        if order.tp_price:
            position._new_tp_order(limit_price=order.tp_price)
        order.parent = position
        return position



import logging
from datetime import datetime
from typing import TYPE_CHECKING, Any, final

from lettrade.account import Account
from lettrade.commander import Commander
from lettrade.data import DataFeed, DataFeeder
from lettrade.exchange import (
    Exchange,
    Execution,
    Order,
    OrderResult,
    Position,
    TradeSide,
)

if TYPE_CHECKING:
    from lettrade.exchange.backtest import BackTestOrder, BackTestPosition

logger = logging.getLogger(__name__)


class Strategy:
    """
    Base class to implement a strategy
    """

    def __init__(
        self,
        feeder: DataFeeder,
        exchange: Exchange,
        account: Account,
        commander: Commander,
        is_optimize: bool = False,
        **kwargs,
    ):
        """_summary_

        Args:
            feeder (DataFeeder): DataFeeder for strategy
            exchange (Exchange): Trading exchange
            account (Account): Account manager
            commander (Commander): Event/Command manager
            is_optimize (bool, optional): flag validate optimize condiction. Defaults to False.

        Raises:
            RuntimeError: Validate valid is_optimize flag

        Example:
            Set custom variable to strategy:
                ```lt = let_backtest(strategy_kwargs=dict(custom_key=custom_value)```

        """
        self.__feeder: DataFeeder = feeder
        self.__exchange: Exchange = exchange
        self.__account: Account = account
        self.__commander: Commander = commander

        self.__datas: list[DataFeed] = self.__feeder.datas
        self.__data: DataFeed = self.__feeder.data

        if is_optimize and self.is_live:
            raise RuntimeError("Optimize a live datafeeder")
        self.__is_optimize: bool = is_optimize

        # Set parameters
        if kwargs:
            logger.info("Update strategy parameters %s", kwargs)
            for k, v in kwargs.items():
                setattr(self, k, v)

    def init(self) -> None:
        """Init strategy variables"""

    def indicators(self, df: DataFeed) -> None:
        """All indicator and signal should implement here to cacheable.
        Because of `lettrade` will cache/pre-load all `DataFeed`.

        To add indicators for a special DataFeed, define a new function name
        `indicators_<DataFeed name>`.

        Usage:
            ```python
            df_eurusd = DataFeed(name="eurusd")
            df_gbpusd = DataFeed(name="gbpusd")

            # This function will load indicators for df_eurusd
            def indicators(self, df: DataFeed):
                df['ema'] = df.i.ema(window=21)

            # This function will load indicators for df_gbpusd
            def indicators_gbpusd(self, df: DataFeed):
                df['ema'] = df.i.ema(window=25)
            ```

        Args:
            df (DataFeed): DataFeed need to load indicators value
        """

    @final
    def _indicators_loader_inject(self):
        for data in self.datas:
            fn_name = f"indicators_{data.name.lower()}"
            if hasattr(self, fn_name):
                fn = getattr(self, fn_name)
            else:
                fn = self.indicators
            object.__setattr__(data, "indicators_load", fn)

    @final
    def _indicators_load(self):
        for data in self.datas:
            data.indicators_load(data)

    @final
    def _start(self):
        self._indicators_loader_inject()
        self._indicators_load()
        self.start(*self.datas)

    def start(self, df: DataFeed, *others: list[DataFeed]) -> None:
        """call after `init()` and before first `next()` is called

        Args:
            df (DataFeed): _description_

        Returns:
            _type_: `None`
        """

    @final
    def _next(self) -> None:
        if self.is_live:
            self._indicators_load()
        self.next(*self.datas)

    def next(self, df: DataFeed, *others: list[DataFeed]) -> None:
        """Next bar event"""

    @final
    def _stop(self) -> None:
        self.stop(*self.datas)

    def stop(self, df: DataFeed, *others: list[DataFeed]) -> None:
        """Call when strategy run completed

        Args:
            df (DataFeed): main data of strategy
        """

    @final
    def _plot(self, *datas: list[DataFeed]) -> dict:
        from lettrade.indicator import indicator_load_plotters
        from lettrade.plot import plot_merge

        config = dict()
        for data in datas:
            plot_merge(config, indicator_load_plotters(data))

        return plot_merge(config, self.plot(*datas))

    def plot(self, df: DataFeed, *others: list[DataFeed]) -> dict:
        """Custom config of plot

        Example:
            ```json
            {
                # Global items is main datafeed item
                "items": [],

                # DataFeed plot config
                "datafeed_name_eurusd": {
                    "items": [
                        {
                            "type": "scatter",
                            "x": dataframe.index,
                            "y": dataframe["close"],
                            "name": "EMA",
                            "mode": "lines",
                            "line": { "color": '#00bad6', "width": 1 },
                        }
                    ]
                },

                # Layout config
                {
                    "hovermode": "x unified"
                }
            }
            ```

        Args:
            df (DataFeed): plot DataFeed

        Returns:
            dict: config
        """
        return dict()

    @final
    def order_buy(
        self,
        size: float | None = None,
        limit: float | None = None,
        stop: float | None = None,
        sl: float | None = None,
        tp: float | None = None,
        tag: str | None = None,
        **kwargs,
    ) -> OrderResult:
        """Place a new long order.

        Args:
            size (float | None, optional): _description_. Defaults to None.
            limit (float | None, optional): _description_. Defaults to None.
            stop (float | None, optional): _description_. Defaults to None.
            sl (float | None, optional): _description_. Defaults to None.
            tp (float | None, optional): _description_. Defaults to None.
            tag (str | None, optional): _description_. Defaults to None.
            **kwargs (dict | None, optional): Extra-parameters send to `Exchange.new_order`

        Returns:
            OrderResult: order result information
        """
        return self.order_place(
            side=TradeSide.Buy,
            size=size,
            limit=limit,
            stop=stop,
            sl=sl,
            tp=tp,
            tag=tag,
            **kwargs,
        )

    buy = order_buy

    @final
    def order_sell(
        self,
        size: float | None = None,
        limit: float | None = None,
        stop: float | None = None,
        sl: float | None = None,
        tp: float | None = None,
        tag: str | None = None,
        **kwargs,
    ) -> OrderResult:
        """Place a new short order.

        Args:
            size (float | None, optional): _description_. Defaults to None.
            limit (float | None, optional): _description_. Defaults to None.
            stop (float | None, optional): _description_. Defaults to None.
            sl (float | None, optional): _description_. Defaults to None.
            tp (float | None, optional): _description_. Defaults to None.
            tag (str | None, optional): _description_. Defaults to None.
            **kwargs (dict | None, optional): Extra-parameters send to `Exchange.new_order`

        Returns:
            OrderResult: order result information
        """
        return self.order_place(
            side=TradeSide.Sell,
            size=size,
            limit=limit,
            stop=stop,
            sl=sl,
            tp=tp,
            tag=tag,
            **kwargs,
        )

    sell = order_sell

    @final
    def order_place(
        self,
        side: TradeSide,
        size: float | None = None,
        limit: float | None = None,
        stop: float | None = None,
        sl: float | None = None,
        tp: float | None = None,
        tag: str | None = None,
        **kwargs,
    ) -> OrderResult:
        """_summary_

        Args:
            side (TradeSide): _description_
            size (float | None, optional): _description_. Defaults to None.
            limit (float | None, optional): _description_. Defaults to None.
            stop (float | None, optional): _description_. Defaults to None.
            sl (float | None, optional): _description_. Defaults to None.
            tp (float | None, optional): _description_. Defaults to None.
            tag (str | None, optional): _description_. Defaults to None.

        Returns:
            OrderResult: _description_
        """
        params = dict(
            size=size,
            limit=limit,
            stop=stop,
            sl=sl,
            tp=tp,
            tag=tag,
            **kwargs,
        )
        params["size"] = side * abs(self.__account.risk(side=side, **params))
        return self.__exchange.new_order(**params)

    @final
    def positions_exit(self, side: TradeSide | None = None):
        for p in list(self.positions.values()):
            if side is not None and p.side != side:
                continue
            p.exit()

    @final
    @property
    def feeder(self) -> DataFeeder:
        """Getter of `DataFeeder`

        Returns:
            DataFeeder: _description_
        """
        return self.__feeder

    @final
    @property
    def exchange(self) -> Exchange:
        """Getter of `Exchange`

        Returns:
            Exchange: _description_
        """
        return self.__exchange

    @final
    @property
    def now(self) -> datetime:
        """Getter of current datetime

        Returns:
            datetime: current datetime of bar
        """
        return self.data.now

    @final
    @property
    def account(self) -> Account:
        """Getter of `Account`

        Returns:
            Account: _description_
        """
        return self.__account

    @final
    @property
    def commander(self) -> Commander:
        """Getter of `Commander`

        Returns:
            Commander: _description_
        """
        return self.__commander

    @final
    @property
    def data(self) -> DataFeed:
        """Getter of main DataFeed

        Returns:
            DataFeed: _description_
        """
        return self.__data

    @final
    @property
    def datas(self) -> list[DataFeed]:
        """Getter of all DataFeed

        Returns:
            list[DataFeed]: _description_
        """
        return self.__datas

    @final
    @property
    def orders(self) -> "dict[str, Order | BackTestOrder]":
        """Getter of `Order` dict

        Returns:
            dict[str, Order]: _description_
        """
        return self.__exchange.orders

    @final
    @property
    def history_orders(self) -> "dict[str, Order | BackTestOrder]":
        """Getter of history `Order` dict

        Returns:
            dict[str, Order]: _description_
        """
        return self.__exchange.history_orders

    @final
    @property
    def positions(self) -> "dict[str, Position | BackTestPosition]":
        """Getter of `Position` dict

        Returns:
            dict[str, Position]: _description_
        """
        return self.__exchange.positions

    @final
    @property
    def history_positions(self) -> "dict[str, Position | BackTestPosition]":
        """Getter of history `Position` dict

        Returns:
            dict[str, Position]: _description_
        """
        return self.__exchange.history_positions

    @final
    @property
    def is_live(self) -> bool:
        """Flag to check strategy is running in live DataFeeder

        Returns:
            bool: _description_
        """
        return self.__feeder.is_continous

    @final
    @property
    def is_backtest(self) -> bool:
        """Flag to check strategy is running in backtest DataFeeder

        Returns:
            bool: _description_
        """
        return not self.is_live

    @final
    @property
    def is_optimize(self) -> bool:
        """Flag to check strategy is running in optimize session

        Returns:
            bool: _description_
        """
        return self.__is_optimize

    # Events
    def on_transactions(self, trans: list[Execution | Order | Position]):
        """Listen for transaction events since `next()` begin

        Args:
            trans (list[Execution  |  Order  |  Position]): _description_
        """
        if hasattr(self, "on_transaction"):
            for tran in trans:
                self.on_transaction(tran)

    def on_executions(self, executions: list[Execution]):
        """Listen for `Execution` event since `next()` begin

        Args:
            executions (list[Execution]): _description_
        """
        if hasattr(self, "on_execution"):
            for execution in executions:
                self.on_execution(execution)

    def on_orders(self, orders: list[Order]):
        """Listen for `Order` event

        Args:
            orders (list[Order]): _description_
        """
        if hasattr(self, "on_order"):
            for order in orders:
                self.on_order(order)

    def on_positions(self, positions: list[Position]):
        """Listen for `Position` event since `next()` begin

        Args:
            positions (list[Position]): _description_
        """
        if hasattr(self, "on_position"):
            for position in positions:
                self.on_position(position)

    def on_notify(self, *args, **kwargs) -> None:
        """Listen for `notify` event since `next()` begin

        Returns:
            _type_: `None`
        """

    # Commander
    @final
    def send_notify(self, msg: str, **kwargs) -> Any:
        """Send message to commander

        Args:
            msg (str): message string

        Returns:
            Any: _description_
        """
        return self.commander.send_message(msg=msg, **kwargs)
